/*
 * Controller2dTop.java:
 *
 * Copyright (C) 2000 John Watson, jwatson@tempusmud.com
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

public class Controller2dTop extends Controller2d {
        public Controller2dTop() {
                super();
        }


        /**
         * this modifies the Point p parameter
         * this handles the flip across the x axis and the translation
         * @param p is changed
         * @return the mutated p
         * 
         */
        public Point3D screenToMap( Point3D p ) {
                p.x /= _view.getZoom();
                p.y /= _view.getZoom();

                Dimension d = _map.getMapSize();

                p.x -= d.width/2;
                p.y = d.height/2 - p.y;

                return p;
        }

        public void mouseMoved( MouseEvent evt ) {
                
                Point scr_p = evt.getPoint();
                
                Point3D p = screenToMap( scr_p );

                _vc.setInfoPointer( p );
                
                boolean contained = _map.mapContains( p );
                
                if ( _vc.isMakingNewSector() && contained ) {
                        
                        Point sp = _vc.isSnapEnabled() ? _vc.getSnapped( p.x, p.y, 0 ) : p;
                        
                        _map.setTransient( true, null, sp );
                        _map.notifyListenersTransient();
                }
        }


        DragAction _dragAction = new DragAction() {
                        public void doAction( Map.Element elem ) {
                                

                                if ( elem instanceof Map.Vertex ) {
                                        _map.moveVertexRelative( (Map.Vertex) elem, diff, true );
                                        return;
                                }
                                if ( elem instanceof Map.LightSource ) {
                                        _map.moveLightRelative( (Map.LightSource) elem, diff, true );
                                        return;
                                }
                                
                                if ( elem instanceof Map.Sector ) {
                                        Map.Vertex[] svs = ( (Map.Sector) elem ).s_vertexes;
                                        
                                        for ( int i = 0; i < svs.length; ++i ) {
                                                if ( ! vertVect.contains( svs[i] ) )
                                                        vertVect.addElement( svs[i] );
                                        }
                                        return;
                                }

                                if ( elem instanceof Map.Line ) {
                                        Map.Vertex[] lvs = ( (Map.Line) elem ).verts;
                                        
                                        if ( ! vertVect.contains( lvs[0] ) )
                                                vertVect.addElement( lvs[0] );
                                        if ( ! vertVect.contains( lvs[1] ) )
                                                vertVect.addElement( lvs[1] );
                                        return;
                                }   
                        }
                };
        
        public void mouseDragged( MouseEvent evt ) {
                
                Point3D p = screenToMap( evt.getPoint() );
                
                _vc.setInfoPointer( p );
                
                int mod = evt.getModifiers();
                
                //
                // middle button down, drag the items
                //

                if ( ( mod & evt.BUTTON2_MASK ) != 0 && _vc.isMode( _vc.MODE_SELECT ) && _map.hasSelections() ) {

                        Point3D diff = new Point3D( p );
                        
                        int gridSize = _vc.getGridSize();
                        int halfGrid = gridSize / 2;
                        
                        Map.Element elem = _map.getLastSelection();
                        
                        if ( elem == null )
                                return;
                        
                        diff.x = p.x - _dragStartPt.x;
                        diff.y = p.y - _dragStartPt.y;
                        
                        //
                        // snap to closest point if needed
                        //

                        if ( _vc.isSnapEnabled() ) {
                                if (diff.x >= halfGrid ) {
                                        diff.x = ( ( diff.x + halfGrid ) / gridSize ) * gridSize;
                                }
                                else if (diff.x <= -halfGrid ) {
                                        diff.x = ( ( diff.x - halfGrid ) / gridSize ) * gridSize;
                                }
                                else {
                                        diff.x = 0;
                                }

                                if ( diff.y >= halfGrid ) {
                                        diff.y = ( ( diff.y + halfGrid ) / gridSize ) * gridSize;
                                }
                                else if ( diff.y <= -halfGrid ) {
                                        diff.y = ( ( diff.y - halfGrid ) / gridSize ) * gridSize;
                                }
                                else {
                                        diff.y = 0;
                                }
                        }

                        
                        //                        System.out.println( " diff=" + diff );
                        if ( diff.x != 0 || diff.y != 0 ) {
                                
                                _dragStartPt.setLocation( _dragStartPt.x + diff.x, _dragStartPt.y + diff.y, 0 );
                                
                                Vector v_vect = new Vector();
                                
                                if ( elem instanceof Map.Vertex ) {
                                        
                                }
                                
                                _dragAction.diff = diff;
                                _dragAction.vertVect.clear();
                                
                                _map.doSelectionActions( _dragAction );
                                
                                for ( Iterator iter = _dragAction.vertVect.iterator(); iter.hasNext(); ) {
                                        _map.moveVertexRelative( (Map.Vertex) iter.next(), diff, true );
                                }

                                _map.notifyListenersFromCachedClip();
                        }
                }
                
                //
                // button 1 down, drag the rubberband
                //

                else if ( ( mod & evt.BUTTON1_MASK ) != 0 && _vc.isMode( _vc.MODE_SELECT ) ) {
                        
                        _map.setRubberBand( Math.min( _dragStartPt.x, p.x ),
                                            Math.min( _dragStartPt.y, p.y ),
                                            - _map.getMapHeightZ() / 2 + 1,
                                            Math.abs( p.x - _dragStartPt.x ),
                                            Math.abs( p.y - _dragStartPt.y ),
                                            _map.getMapHeightZ() - 1 );
                        
                        _view.setCursor( Cursor.getPredefinedCursor( Cursor.HAND_CURSOR ) );
                        _rubberBanding = true;
                }

                // drag the clip if needed

                else if ( _map._transClip != null && _vc.isMode( _vc.MODE_SET_CLIP ) ) {
                        if ( ( mod & evt.BUTTON1_MASK ) != 0 ) {
                                
                                if ( p.x < ( _map._transClip.x + _map._transClip.width ) ) {
                                        _map._transClip.width -= p.x - _map._transClip.x;
                                        _map._transClip.x = p.x;
                                }
                                if ( p.y < ( _map._transClip.y + _map._transClip.height ) ) {
                                        _map._transClip.height -= p.y - _map._transClip.y;
                                        _map._transClip.y = p.y;
                                }

                                _map.notifyListenersTransient();
                                
                        }

                        if ( ( mod & evt.BUTTON3_MASK ) != 0 ) {

                                if ( p.x > _map._transClip.x ) {
                                        _map._transClip.width = p.x - _map._transClip.x;
                                }
                                if ( p.y > _map._transClip.y ) {
                                        _map._transClip.height = p.y - _map._transClip.y;
                                }

                                _map.notifyListenersTransient();

                        }
                }
        }

        public void mousePressed( MouseEvent evt ) {
                Point3D p = screenToMap( evt.getPoint() );

                int mod = evt.getModifiers();
                                
                _dragStartPt.setLocation( p );

                if ( _vc.isMode( _vc.MODE_SELECT ) && ( mod & evt.BUTTON2_MASK ) != 0 ) {
                        if ( _map.getLastSelection() != null )
                                _view.setCursor( Cursor.getPredefinedCursor( Cursor.MOVE_CURSOR ) );
                        return;
                }
                
                if ( _vc.isMode( _vc.MODE_NEW_LIGHT ) ) {
                        Point3D snapped = _vc.isSnapEnabled() ? _vc.getSnapped( p ) : p;

                        if ( _map.findLight( snapped ) != null ) {
                                Toolkit.getDefaultToolkit().beep();
                                return;
                        }

                        Map.LightSource light = new Map.LightSource();
                        light.pos.setLocation( snapped.x,
                                               snapped.y,
                                               0 );
                        light.r = 255;
                        light.g = 255;
                        light.b = 255;
                        light.intensity = 256;

                        _map.addLight( light );
                        return;
                }

                if ( _vc.isMode( _vc.MODE_NEW_SECTOR ) ) {
                        
                        //
                        // continue making a sector
                        //

                        if ( _vc.isMakingNewSector() ) {

                                //
                                // cancel
                                //

                                if ( ( mod & evt.BUTTON2_MASK ) != 0 ) {
                                        _vc.cancelNewSector();
                                }
                                
                                //
                                // close the sector
                                //

                                else if ( ( mod & evt.BUTTON3_MASK ) != 0 ) {
                                        if ( _map.numPtsInProg <= 2 ) {
                                                System.out.println( " polygon needs 3 verts or more" );
                                        }
                                        else { 
                                                Map.Sector s = new Map.Sector( _map.ptsInProg, _map.numPtsInProg );
                                                _map.addSector( s );
                                                _vc.cancelNewSector();
                                        }
                                }

                                //
                                // add a point
                                //

                                else if ( ( mod & evt.BUTTON1_MASK ) != 0 ) {

                                        Point snapped = _vc.isSnapEnabled() ? _vc.getSnapped( p.x, p.y, 0 ) : p;
                                        
                                        //
                                        // prevent the user from placing a vert on an existing vert
                                        // of this new sector (excluding the first )
                                        //

                                        for ( int i = 1; i < _map.numPtsInProg; ++i ) {
                                                if ( _map.ptsInProg[0][i] == snapped.x &&
                                                     _map.ptsInProg[1][i] == snapped.y ) {
                                                        System.out.println( " don't put a vertex there, dumbass." );
                                                        return;
                                                }
                                        }

                                        //
                                        // if this point is equal to the startpoint, close the sector now
                                        //
                                        
                                        if ( _map.ptsInProg[0][0] == snapped.x &&
                                             _map.ptsInProg[1][0] == snapped.y ) {
                                                System.out.println( " forcing sector close" );
                                                if ( _map.numPtsInProg <= 2 ) {
                                                        System.out.println( " polygon needs 3 verts or more" );
                                                }
                                                else {
                                                        Map.Sector s = new Map.Sector( _map.ptsInProg, _map.numPtsInProg );
                                                        _map.addSector( s );
                                                        _vc.cancelNewSector(); 
                                                }
                                        }
                                        else {
                                                _map.addInProgress( snapped );
                                                _map.setTransient( true, snapped, snapped );

                                                Polygon poly = new Polygon( _map.ptsInProg[0],
                                                                            _map.ptsInProg[1], 
                                                                            _map.numPtsInProg );
                                                _map.notifyListeners( poly.getBounds() );
                                        }
                                }
                        }
                        
                        //
                        // start a new sector
                        //

                        else {

                                if ( ( mod & evt.BUTTON1_MASK ) != 0 ) {
                                        Point snapped = _vc.isSnapEnabled() ? _vc.getSnapped( p.x, p.y, 0 ) : p;
                                        
                                        _vc.startNewSector( snapped );
                                }
                        }
                }

                // set clip
                
                if ( _vc.isMode( _vc.MODE_SET_CLIP ) ) { 
                        
                        if ( _map._transClip == null )
                                _map._transClip = new Rectangle( _map._clip );

                        Rectangle old = new Rectangle( _map._transClip );

                        if ( ( mod & evt.BUTTON1_MASK ) != 0 ) {
                                
                                if ( p.x < ( _map._transClip.x + _map._transClip.width ) ) {
                                        _map._transClip.width -= p.x - _map._transClip.x;
                                        _map._transClip.x = p.x;
                                }
                                if ( p.y < ( _map._transClip.y + _map._transClip.height ) ) {
                                        _map._transClip.height -= p.y - _map._transClip.y;
                                        _map._transClip.y = p.y;
                                }
                                
                                _view.setCursor( Cursor.getPredefinedCursor( Cursor.SW_RESIZE_CURSOR ) );

                                _map.notifyListenersTransient();
                                
                        }

                        if ( ( mod & evt.BUTTON3_MASK ) != 0 ) {

                                if ( p.x > _map._transClip.x ) {
                                        _map._transClip.width = p.x - _map._transClip.x;
                                }
                                if ( p.y > _map._transClip.y ) {
                                        _map._transClip.height = p.y - _map._transClip.y;
                                }

                                _view.setCursor( Cursor.getPredefinedCursor( Cursor.NE_RESIZE_CURSOR ) );

                                _map.notifyListenersTransient();

                        }
                }

        }
        
        public void mouseReleased( MouseEvent evt ) {

                int mod = evt.getModifiers();

                if ( _rubberBanding ) {

                        final Rectangle big_rect = new Rectangle( _map._rubberBand[0].x,
                                                                  _map._rubberBand[0].y,
                                                                  _map._rubberBand[1].x,
                                                                  _map._rubberBand[1].y );

                        final Rectangle rect = big_rect.intersection( _map._clip );

                        _map.unsetRubberBand();
                        _rubberBanding = false;
                        _view.setCursor( null );

                        if ( ! evt.isControlDown() && ! evt.isShiftDown() ) {
                                _map.clearSelections( true );
                        }
                        
                        final Map.Element elem = _map.getLastSelection();

                        //
                        // select all sects in rubber band area
                        //

                        if ( ( elem == null && _vc.isAutoSelect( _vc.AUTO_SECTORS ) ) ||
                             elem instanceof Map.Sector ) {
                                
                                for ( Iterator iter = _map.sectors.iterator(); iter.hasNext(); ) {
                                        
                                        Map.Sector s = (Map.Sector) iter.next();
                                        
                                        if ( s.getPolygon().intersects( rect ) ) {
                                                if ( evt.isControlDown() ) {
                                                        _map.toggleSelection( s, true );
                                                }
                                                else {
                                                        _map.addSelection( s, true );
                                                }
                                        }
                                }
                        }
                        
                        //
                        // select all lines in rubber band area
                        //
                        
                        else if ( ( elem == null && _vc.isAutoSelect( _vc.AUTO_LINES ) ) ||
                                  elem instanceof Map.Line ) {

                                rect.grow( 2, 2 );
                                
                                for ( Iterator iter = _map.lines.iterator(); iter.hasNext(); ) {
                                        Map.Line line = (Map.Line) iter.next();
                                        
                                        if ( rect.contains( line.verts[0].x, line.verts[0].y ) &&
                                             rect.contains( line.verts[1].x, line.verts[1].y ) ) {
                                                if ( evt.isControlDown() ) {
                                                        _map.toggleSelection( line, true );
                                                }
                                                else {
                                                        _map.addSelection( line, true );
                                                }
                                        }
                                }
                        }
                        
                        //
                        // select all verts in rubber band area
                        //

                        else if ( ( elem == null && _vc.isAutoSelect( _vc.AUTO_VERTS ) ) ||
                                  elem instanceof Map.Vertex ) {
                                rect.grow( 4, 4 );
                                
                                for ( Iterator iter = _map.vertexes.iterator(); iter.hasNext(); ) {
                                        Map.Vertex v = (Map.Vertex) iter.next();
                                        
                                        if ( rect.contains( v.x, v.y ) ) {
                                                if ( evt.isControlDown() ) {
                                                        _map.toggleSelection( v, true );
                                                }
                                                else {
                                                        _map.addSelection( v, true );
                                                }    
                                        }
                                }
                        }

                        //
                        // select all lights in rubber band area
                        //

                        else if ( ( elem == null && _vc.isAutoSelect( _vc.AUTO_LIGHTS ) ) ||
                                  elem instanceof Map.LightSource ) {
                                
                                rect.grow( 16, 16 );
                                
                                for ( Iterator iter = _map.lights.iterator(); iter.hasNext(); ) {
                                        Map.LightSource light = (Map.LightSource) iter.next();
                                        
                                        if ( rect.contains( light.pos.x, light.pos.y ) ) {
                                                if ( evt.isControlDown() ) {
                                                        _map.toggleSelection( light, true );
                                                }
                                                else {
                                                        _map.addSelection( light, true );
                                                }    
                                        }
                                }

                        }

                        //
                        // select all entities in rubber band area
                        //

                        else if ( ( elem == null &&_vc.isAutoSelect( _vc.AUTO_ENTITIES ) ) ) {
                                
                        }

                        // see if any selections on the map actually changed
                        
                        if ( _map._isClipCached ) {
                                _map.addCacheClip( big_rect );
                                _map.notifyListenersFromCachedClip();

                        }
                        else {
                                _map.notifyListenersTransient();
                        }
                }
                
                else if ( _vc.isMode( _vc.MODE_SELECT ) && ( mod & evt.BUTTON2_MASK ) != 0 ) {
                        _view.setCursor( null );
                
                }
        
                // set clip to match transient clip
        
                else if ( _vc.isMode( _vc.MODE_SET_CLIP ) ) { 
                        
                        if ( _map._transClip != null ) {

                                _view.setCursor( Cursor.getPredefinedCursor( Cursor.HAND_CURSOR ) );

                                _map.setUserClip( _map._transClip );
                                _map._transClip = null;
                        }
                }
        
        }

        public void mouseClicked( MouseEvent evt ) {
                Object o = evt.getSource();
                
                if ( o instanceof Component ) {
                        ( (Component) o ).requestFocus();
                }

                Point p = screenToMap( evt.getPoint() );

                int mod = evt.getModifiers();

                boolean control_down = ( mod & evt.CTRL_MASK ) != 0;
                boolean shift_down   = ( mod & evt.SHIFT_MASK ) != 0;
                boolean single_select = control_down == false && shift_down == false;

                if ( _vc.isMode( _vc.MODE_SELECT ) && ( mod & evt.BUTTON1_MASK ) != 0 ) {
                        
                        if ( _map._clip.contains( p ) == false )
                                return;

                        Map.Element elem = null;

                        //
                        // look for appropriate lights to select
                        //

                        elem = _map.getLastSelection();

                        if ( single_select ||
                             elem == null || elem instanceof Map.LightSource ) {
                                
                                for ( Iterator iter = _map.lights.iterator(); iter.hasNext(); ) {
                                
                                        Map.LightSource light = (Map.LightSource) iter.next();
                                        
                                        if ( VMath.inRange( light.pos.x, light.pos.y, p.x, p.y, 16 ) ) {
                                                        
                                                if ( control_down ) {
                                                        _map.toggleSelection( light, false );
                                                }
                                                else if ( shift_down ) {
                                                        _map.addSelection( light, false );
                                                }
                                                else {
                                                        _map.setSelection( light, false );
                                                }
                                                return;
                                        }
                                }
                                
                        }
                        
                        //
                        // look for appropriate verts to select
                        //

                        
                        elem = _map.getLastSelection();
                        
                        if ( single_select ||
                             elem == null || elem instanceof Map.Vertex ) {
                                for ( Iterator iter = _map.vertexes.iterator(); iter.hasNext(); ) {
                                        Map.Vertex v = (Map.Vertex) iter.next();
                                        
                                        if ( VMath.inRange( v.x, v.y, p.x, p.y, 4 ) ) {
                                                if ( control_down ) {
                                                        _map.toggleSelection( v, false );
                                                }
                                                else if ( shift_down ) {
                                                        _map.addSelection( v, false );
                                                }
                                                else {
                                                        _map.setSelection( v, false );
                                                }
                                                
                                                return;
                                        }
                                }
                        }

                        //
                        // look for appropriate lines to select
                        //

                        elem = _map.getLastSelection();

                        if ( single_select ||
                             elem == null || elem instanceof Map.Line ) {
                        
                                for ( Iterator iter = _map.lines.iterator(); iter.hasNext(); ) {
                                        Map.Line line = (Map.Line) iter.next();
                                
                                        if ( VMath.distFromLine( p.x, p.y, 
                                                                 line.verts[0].x, line.verts[0].y, 
                                                                 line.verts[1].x, line.verts[1].y, true ) <= 
                                             ( elem != null ? 8 : 4 ) ) {
                                                if ( control_down ) {
                                                        _map.toggleSelection( line, false );
                                                }
                                                else if ( shift_down ) {
                                                        _map.addSelection( line, false );
                                                }
                                                else {
                                                        _map.setSelection( line, false );
                                                }
                                                return;
                                        }
                                }
                        }
                        
                        //
                        // look for appropriate sectors to select
                        //

                        elem = _map.getLastSelection();

                        if ( single_select ||
                             elem == null || elem instanceof Map.Sector ) {
                                
                                for ( Iterator iter = _map.sectors.iterator(); iter.hasNext(); ) {
                                
                                        Map.Sector s = (Map.Sector) iter.next();
                                
                                        if ( s.getPolygon().contains( p ) ) {
                                        
                                                if ( control_down ) {
                                                        _map.toggleSelection( s, false );
                                                }
                                                else if ( shift_down ) {
                                                        _map.addSelection( s, false );
                                                }
                                                else {
                                                        _map.setSelection( s, false );
                                                }
                                                return;
                                        }
                                }
                                
                        }

                        if ( ! shift_down && ! control_down )
                                _map.clearSelections( false );
                        
                        return;
                }
        }
}
