/*
 * Map.java:
 *
 * Copyright (C) 2000 John Watson, jwatson@tempusmud.com
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */
import java.util.*;
import java.awt.*;

import java.io.*;

public class Map implements Serializable {

        private int LAST_SECTOR_NUM = 0;

        public static final String JSFMAP_EXT = "jsf";
        public static final String DGMAP_EXT  = "zon";

        public static final long serialVersionUID = 1;

        String name = null;
        
        Vector vertexes     = new Vector();
        Vector lines        = new Vector();
        Vector sectors      = new Vector();
        Vector lights       = new Vector();
        Vector sectSpecials = new Vector();
        Vector scripts      = new Vector();

        transient boolean drawTransient = false;
        
        transient Point[] tranPts = new Point[] {
                new Point(),
                new Point(),
        };
        
               
        transient int[][] ptsInProg = new int[2][100];
        transient int numPtsInProg = 0;
        
        transient Point3D[] _rubberBand = null;

        transient Rectangle _clip = new Rectangle();
        transient Rectangle _transClip = null;

        transient Rectangle _internalNotifyClip = new Rectangle();

        public static final int DEF_FLOOR_TEX   = 0;
        public static final int DEF_CEILING_TEX = 1;
        public static final int DEF_WALL_TEX    = 2;
        public static final int NUM_DEF_TEXS    = 3;
        
        transient int preferredLineSide = 0;
        
        transient private int[] _defaultTexIndexes = new int[ NUM_DEF_TEXS ];
        private String[] _defaultTexs = new String[ NUM_DEF_TEXS ];

        transient private Vector _selected = new Vector();
        
        transient Vector listeners = new Vector();

        transient Vector scriptListeners = new Vector();

        Dimension _mapSize = new Dimension();
        Rectangle _mapRect = new Rectangle();

        private Map() {}

        public Map( String name, int x, int y, 
                    String defFloorTex, 
                    String defCeilingTex, 
                    String defWallTex ) {
                this.name = name;
                _mapSize.setSize( x, y );
                _mapRect.setBounds( - x/2, - y/2, x, y );
                _clip.setBounds( _mapRect );
                _transClip = null;

                setDefaultTex( DEF_FLOOR_TEX,   defFloorTex );
                setDefaultTex( DEF_CEILING_TEX, defCeilingTex );
                setDefaultTex( DEF_WALL_TEX,    defWallTex );
        }
        
        public void setMapSize( int x, int y ) {
                _mapSize.setSize( x, y );
                _mapRect.setBounds( - x/2, - y/2, x, y );
                _clip.setBounds( _mapRect );
                _transClip = null;
                notifyListenersSize();
        }

        public int getMapHeightZ() {
                return 4096;//_mapSize.height;
        }

        public Dimension getMapSize() {
                return _mapSize;
        }

        public Rectangle getMapRect() {
                return _mapRect;
        }

        public boolean mapContains( Point p ) {
                return _mapRect.contains( p );
        }
        
        //
        // methods for manipulating the selected element(s)
        //
        
        public boolean addSelection( Element elem, boolean cache ) {

                //
                // don't allow mixed-type selections!
                //

                Element last = getLastSelection();
                if ( last != null && last.getClass() != elem.getClass() )
                        return false;

                if ( ! _selected.contains( elem ) ) {
                        _selected.addElement( elem );
                        elem.isSelected = true;

                        if ( cache ) {
                                addCacheClip( elem.getBounds( 0 ) );
                        }
                        else {
                                notifyListeners( elem.getBounds( 0 ) );
                        }

                        return true;
                }
                return false;
        }

        public boolean isSelected( Element elem ) {
                return _selected.contains( elem );
        }

        public boolean removeSelection( Element elem, boolean cache ) {
                boolean b = _selected.removeElement( elem );
                if ( b ) {
                        elem.isSelected = false;
                        if ( cache )
                                addCacheClip( elem.getBounds( 0 ) );
                        else
                                notifyListeners( elem.getBounds( 0 ) );
                }
                return b;
        }

        public boolean toggleSelection( Element elem, boolean cache ) {
                if ( isSelected( elem ) )
                        return removeSelection( elem, cache );
                else
                        return addSelection( elem, cache );
        }

        public void clearSelections( boolean cache ) {
                
                for ( Iterator iter = _selected.iterator(); iter.hasNext(); ) {
                        Element elem = (Element) iter.next();
                        
                        elem.isSelected = false;
                        addCacheClip( elem.getBounds( 0 ) );
                        iter.remove();
                }
                
                if ( cache == false )
                        notifyListenersFromCachedClip();
        }

        public boolean setSelection( Element elem, boolean cache ) {
                clearSelections( cache );
                return addSelection( elem, cache );
        }

        public int getNumSelections() {
                return _selected.size();
        }

        public Element getLastSelection() {
                if ( _selected.size() > 0 )
                        return (Element) _selected.lastElement();
                return null;
        }

        public Element getFirstSelection() {
                if ( _selected.size() > 0 )
                        return (Element) _selected.firstElement();
                return null;
        }
        
        public Vertex getLastSelectedVertex() {
                Object o = _selected.lastElement();
                if ( o instanceof Vertex )
                        return (Vertex) o;
                return null;
        }
        
        public void doSelectionActions( ElementAction act ) {
                for ( Iterator iter = _selected.iterator(); iter.hasNext(); ) {
                        act.doAction( (Element) iter.next() );
                }
        }
        
        public boolean hasTypeSelected( Class c ) {
                for ( Iterator iter = _selected.iterator(); iter.hasNext(); ) {
                        if ( iter.next().getClass() == c )
                                return true;
                }
                return false;
        }

        public void setUserClip( Rectangle clip ) {
                _clip = clip;
                for ( Iterator iter = _selected.iterator(); iter.hasNext(); ) {
                        Element elem = (Element) iter.next();

                        if ( _clip.intersects( elem.getSubclassBounds() ) == false ) {
                                elem.isSelected = false;
                                iter.remove();
                        }
                }

                notifyListeners( null );
        }
                
                

        /**
         * find the centerpoint of all selected elements
         * the centerpoint is the center of the bounding box which contains
         * all the other centers
         */

        public Point3D findSelectionsCenter() {
                Point3D p = new Point3D();
                
                if ( _selected.size() <= 0 ) {
                        return new Point3D( _clip.x + _clip.width / 2,
                                            _clip.y + _clip.height / 2,
                                            0 );
                }

                int min_z = _mapSize.height / 2;
                int max_z = - min_z;
                
                Rectangle rect = null;
                
                for ( Iterator iter = _selected.iterator(); iter.hasNext(); ) {
                        Element elem = (Element) iter.next();
                        
                        Point3D ep = elem.getCenterPoint3D();
                        
                        if ( rect == null ) 
                                rect = new Rectangle( ep );
                        else
                                rect.add( ep );
                        
                        max_z = Math.max( max_z, ep.z );
                        min_z = Math.min( min_z, ep.z );
                }


                p.x = rect.x + rect.width / 2;
                p.y = rect.y + rect.height / 2;
                p.z = ( min_z + max_z ) / 2;

                return p;

        }

        /**
         * find the bounding rect which contains all selections
         * @return null if no selections,
         */

        public Rectangle findSelectionsBounding() {
                Rectangle rect = null;
                
                for ( Iterator iter = _selected.iterator(); iter.hasNext(); ) {
                        Element elem = (Element) iter.next();
                        
                        if ( rect == null )
                                rect = new Rectangle( elem.getSubclassBounds() );
                        else
                                rect.add( elem.getSubclassBounds() );
                }

                return rect;

        }
        
        
        public int numSelections() {
                return _selected.size();
        }
        public boolean hasSelections() {
                return ( _selected.size() > 0 );
        }

        //
        // methods for manipulating the default textures
        //

        public int getDefaultTex( int which ) {
                return _defaultTexIndexes[ which ];
        }

        public void setDefaultTex( int which, int index ) {
                _defaultTexIndexes[ which ] = index;
                _defaultTexs[ which ] = TextureManager.getTexname( index );
        }

        public void setDefaultTex( int which, String str ) {
                _defaultTexIndexes[ which ] = TextureManager.findTexture( str );
                if ( _defaultTexIndexes[ which ] > 0 )
                        _defaultTexs[ which ] = str;
                else
                        _defaultTexs[ which ] = null;
        }

        void clear() {
                clearSelections( true );
                sectors.removeAllElements();
                vertexes.removeAllElements();
                lines.removeAllElements();
                numPtsInProg = 0;
                drawTransient = false;
                notifyListeners( null );
        }
        
        public void save( String filename ) {
        }

        synchronized public void addInProgress( Point p ) {
                ptsInProg[0][ numPtsInProg ] = p.x;
                ptsInProg[1][ numPtsInProg ] = p.y;
                ++numPtsInProg;
        }

        public void setTransient( boolean state, Point p0, Point p1 ) {
                drawTransient = state;

                if ( state ) {
                        
                        if ( p0 != null ) {
                                tranPts[0].setLocation( p0 );
                        }
                        if ( p1 != null ) {
                                tranPts[1].setLocation( p1 );
                        }
                }
        }
        
        public void unsetRubberBand() {
                _rubberBand = null;
        }

        public void setRubberBand( int ox, int oy, int oz,
                                   int ex, int ey, int ez ) {

                if ( _rubberBand == null ) {
                        _rubberBand = new Point3D[2];
                        _rubberBand[0] = new Point3D( ox, oy, oz );
                        _rubberBand[1] = new Point3D( ex, ey, ez );
                }
                else {
                        _rubberBand[0].setLocation( ox, oy, oz );
                        _rubberBand[1].setLocation( ex, ey, ez );
                }
                        
                notifyListenersTransient();
        }
        
        void addCacheClip( Rectangle clip ) {
                if ( ! clip.isEmpty() ) {
                        if ( ! _isClipCached ) {
                                _internalNotifyClip.setBounds( clip );
                                _isClipCached = true;
                        }
                        else {
                                _internalNotifyClip.add( clip );
                        }
                }
        }

        public boolean _isClipCached = false;

        /**
         * notify map listeners based on the interally maintained
         * aggregate cache of map modification clip rectangles
         *
         */
        
        public void notifyListenersFromCachedClip() {
                if ( _isClipCached == true ) {
                        notifyListeners( _internalNotifyClip );
                        _isClipCached = false;
                }
        }
        
        public void notifyListeners( Rectangle clip ) {
                for ( Enumeration enum = listeners.elements(); enum.hasMoreElements(); ) {
                        ( (Listener) enum.nextElement() ).mapChanged( clip );
                }
        }

        public void notifyListenersTransient() {
                for ( Enumeration enum = listeners.elements(); enum.hasMoreElements(); ) {
                        ( (Listener) enum.nextElement() ).mapTransientChanged();
                }
        }

        public void notifyListenersSize() {
                for ( Enumeration enum = listeners.elements(); enum.hasMoreElements(); ) {
                        ( (Listener) enum.nextElement() ).mapSizeChanged();
                }
        }
        
        public void notifyScriptListeners() {
                for ( Iterator iter = scriptListeners.iterator(); iter.hasNext(); ) {
                        ( (ScriptListener) iter.next() ).scriptsChanged();
                }
        } 

        public void addLight( LightSource light ) {
                lights.addElement( light );
                notifyListeners( light.getBounds( 4 ) );
        }
        
        public void addSector( Sector s ) {
                
                Vertex last_v = null;
                
                int[] wallTexs = new int[] { 
                        getDefaultTex( DEF_WALL_TEX ),
                        getDefaultTex( DEF_WALL_TEX ),
                        getDefaultTex( DEF_WALL_TEX )
                };
                
                for ( int i = 0; i < s.s_vertexes.length; ++i ) {
                        Vertex v = createVertex( s.s_vertexes[i].x, s.s_vertexes[i].y );
                        s.s_vertexes[i] = v;

                        if ( last_v != null ) {
                                Line line = createLine( last_v, v );

                                if ( line.line_sides[0] == null ) {
                                        line.line_sides[0] = new Side( line, s, wallTexs );
                                }
                                else if ( line.line_sides[1] == null ) {
                                        line.line_sides[1] = new Side( line, s, wallTexs );
                                }
                                else {
                                        System.out.println( " SECTOR ADD ERROR" );
                                        purgeVertexes();
                                        purgeLines();
                                        return;
                                }
                                
                        }

                        last_v = v;
                }
                
                Line line = createLine( last_v, s.s_vertexes[0] );

                if ( line.line_sides[0] == null ) {
                        line.line_sides[0] = new Side( line, s, wallTexs );
                }
                else if ( line.line_sides[1] == null ) {
                        line.line_sides[1] = new Side( line, s, wallTexs );
                }
                else {
                        System.out.println( " SECTOR ADD ERROR" );
                        purgeVertexes();
                        purgeLines();
                        return;
                }
                
                sectors.addElement( s );
                s._map = this;
                s.setTex( s.FLOOR_TEX, getDefaultTex( DEF_FLOOR_TEX ) );
                s.setTex( s.CEILING_TEX, getDefaultTex( DEF_CEILING_TEX ) );
                s.num = ++LAST_SECTOR_NUM;
        }
        
        private void removeSector( Sector s ) {
                sectors.removeElement( s );

                for ( Iterator iter = lines.iterator(); iter.hasNext(); ) {
                        Line line = (Line) iter.next();     
                        
                        if( line.line_sides[0] != null && line.line_sides[0].sect == s ) {
                                line.line_sides[0] = null;
                        }
                        if( line.line_sides[1] != null && line.line_sides[1].sect == s ) {
                                line.line_sides[1] = null;
                        }

                        if ( line.line_sides[0] == null ) {

                                //
                                // left a backside with no frontside, flip the line
                                //
                                
                                if ( line.line_sides[1] != null ) {
                                        line.flipLine();
                                }
                                
                                //
                                // this line has no sector, get rid of it
                                //

                                else {
                                        removeSelection( line, true );
                                        iter.remove();
                                }
                        }
                }
                purgeVertexes();
                purgeLines();
                notifyListenersFromCachedClip();
        }
        
        public void removeElement( Element e ) {

                removeSelection( e, true );

                if ( e instanceof Sector ) {
                        removeSector( (Sector) e );
                }
                else if ( e instanceof Line ) {
                        Line line = (Line) e;
                        line.verts[1].x = line.verts[0].x;
                        line.verts[1].y = line.verts[0].y;
                        mergeVertexes( line.verts[0] );
                }
                else if ( e instanceof Vertex ) {
                        
                }
                else if ( e instanceof LightSource ) {
                        lights.removeElement( e );
                }

                //                notifyListeners( e.getBounds( 1 ) );
                notifyListenersFromCachedClip();

        }

        private void moveVertexInternal( Vertex v, int x, int y ) {
                v.x = x;
                v.y = y;
        }

        public void moveVertexRelative( Vertex v, Point diff, boolean cache ) {
                Rectangle clip = v.getBounds( 0 );
                
                moveVertexInternal( v, 
                                    v.x + diff.x, 
                                    v.y + diff.y );
                clip.add( v.getBounds( 0 ) );
                
                for ( Iterator iter = lines.iterator(); iter.hasNext(); ) {
                        Line line = (Line) iter.next();
                        if ( line.hasVert( v ) )
                                clip.add( line.getBounds( 1 ) );
                }
                
                if ( cache )
                        addCacheClip( clip );
                else
                        notifyListeners( clip );
        }

        public void moveLightRelative( LightSource light, Point3D diff, boolean cache ) {
                Rectangle clip = light.getBounds( 0 );
                
                light.pos.add( diff );
                
                clip.add( light.getBounds( 0 ) );
                
                if ( cache )
                        addCacheClip( clip );
                else
                        notifyListeners( clip );
        }

        public boolean moveSector( Sector s, Point dist, boolean cache ) {

                if ( dist.x == 0 && dist.y == 0 )
                        return false;
                
                Rectangle clip = s.getBounds( 1 );
                
                for ( int i = 0; i < s.s_vertexes.length; ++i ) {
                        moveVertexInternal( s.s_vertexes[i], 
                                            s.s_vertexes[i].x + dist.x, 
                                            s.s_vertexes[i].y + dist.y );
                }
                
                clip.add( s.getBounds( 1 ) );

                if ( cache )
                        addCacheClip( clip );
                else
                        notifyListeners( clip );
                
                return true;
        }

        public boolean moveLine( Line line, Point dist ) {
                
                if ( dist.x == 0 && dist.y == 0 )
                        return false;
                
                Rectangle clip = line.getBounds( 1 );
                clip.add( new Point( line.verts[1].x, line.verts[1].y ) );

                moveVertexInternal( line.verts[0], 
                                    line.verts[0].x + dist.x, 
                                    line.verts[0].y + dist.y );

                moveVertexInternal( line.verts[1], 
                                    line.verts[1].x + dist.x, 
                                    line.verts[1].y + dist.y );
                
                
                for ( Iterator iter = lines.iterator(); iter.hasNext(); ) {
                        Line aline = (Line) iter.next();
                        if ( aline.hasVert( line.verts[0] ) )
                                clip.add( aline.getBounds( 1 ) );
                        else if ( aline.hasVert( line.verts[1] ) )
                                clip.add( aline.getBounds( 1 ) );
                }

                notifyListeners( clip );
                return true;  
        }
        
        public void mergeVertexes( Vertex v ) {
                int index = vertexes.indexOf( v );
                

                if ( index == -1 ) {
                        System.out.println( " error in mergeVertexes" );
                        return;
                }
                
                Rectangle clip = new Rectangle( v.x, v.y, 0, 0 );
                
                Vertex target = (Vertex) vertexes.elementAt( index );
                
                int otherindex = index;

                boolean found = false;
                
                Vector swappedLines = new Vector();

                while ( ( otherindex = vertexes.indexOf( v, otherindex + 1 ) ) != -1 ) {

                        found = true;
                        Vertex otherv = (Vertex) vertexes.elementAt( otherindex );
                        
                        //
                        // TODO: make sure the remaining vert is selected if needed
                        //

                        //
                        // fixup the sectors vert references, resizing the sectors if needed
                        //

                        for ( Iterator iter = sectors.iterator(); iter.hasNext(); ) {
                                Sector s = (Sector) iter.next();
                                s.swapVertexRefs( otherv, target );
                        }
                        
                        //
                        // remove all the lines that connect the merging verts
                        //
                        
                        Line connectingLine = findLine( v, otherv );
                        lines.removeElement( connectingLine );

                        //
                        // make sure lines that will be removed will preserve their side
                        //
                        
                        for ( Iterator iter = lines.iterator(); iter.hasNext(); ) {
                                Line line = (Line) iter.next();
                                
                                if ( line.verts[0] != target  &&
                                     line.verts[1] != target ) {
                                        continue;
                                }
                                
                                if ( line.verts[0] != target )
                                        clip.add( new Point( line.verts[0].x, line.verts[0].y ) );
                                else
                                        clip.add( new Point( line.verts[1].x, line.verts[1].y ) );
                                
                                if ( line.line_sides[1] != null )
                                        continue;

                                for ( Iterator biter = lines.iterator(); biter.hasNext(); ) {
                                        Line bline = (Line) biter.next();
                                        
                                        if ( bline == line )
                                                continue;
                                        
                                        //
                                        // a line with the same vertex
                                        //
                                        
                                        if ( bline.equals( line ) ) {
                                                
                                                //
                                                // only blines pointing in opposite direction than line are correct
                                                // 

                                                if ( bline.verts[0] != line.verts[0] ) {
                                                        line.line_sides[1] = bline.line_sides[0];
                                                }
                                                
                                                //
                                                // lines pointing in the same dir is an error but possible, try to fix it up
                                                //
                                                
                                                else {
                                                        line.line_sides[1] = bline.line_sides[1];
                                                }
                                        }
                                }
                        }

                        
                        //
                        // fixup the lines vert references
                        //

                        for ( Iterator iter = lines.iterator(); iter.hasNext(); ) {
                                Line line = (Line) iter.next();
                                
                                if ( swappedLines.contains( line ) )
                                        iter.remove();
                                else {
                                        line.swapVertexRefs( otherv, target );
                                        swappedLines.addElement( line );
                                }
                        }
                }

                if ( ! found ) {
                        System.out.println( " can't merge a vert with itself, idiot." );
                        return;
                }
                
                purgeSectors();
                purgeVertexes();
                purgeLines();
                
                addCacheClip( clip );
                notifyListenersFromCachedClip();
        }

        private void purgeVertexes() {
                
                int num = 0;
                
                for ( Iterator iter = vertexes.iterator(); iter.hasNext(); ) {
                        Vertex v = (Vertex) iter.next();
                        boolean found = false;
                        
                        for ( Iterator siter = sectors.iterator(); siter.hasNext(); ) {
                                Sector s = (Sector) siter.next();
                                if ( s.hasVertex( v ) ) {
                                        found = true;
                                        break;
                                }
                        }

                        if ( ! found ) {
                                removeSelection( v, true );
                                iter.remove();
                                ++num;
                        }
                }
        }

        /**
         * get rid of lines with:
         * 1. verts that are no longer part of the global vertex list
         * 2. lines with no sides
         * 
         */

        private void purgeLines() {
                for ( Iterator iter = lines.iterator(); iter.hasNext(); ) {
                        Line line = (Line) iter.next();

                        if ( ! vertexes.contains( line.verts[0] ) ||
                             ! vertexes.contains( line.verts[1] ) ) {
                                removeSelection( line, true );
                                iter.remove();
                        }
                }
        }

        private int purgeSectors() {
                
                int num =0;
                for ( Iterator iter = sectors.iterator(); iter.hasNext(); ) {
                        Sector s = (Sector) iter.next();
                        if ( s.s_vertexes.length < 3 ) {
                                removeSelection( s, true );
                                iter.remove();
                                ++num;
                        }
                }

                return num;
        }

        private Vertex findVertex( int x, int y ) {
                Vertex mv = new Vertex( x, y );
                int index = vertexes.indexOf( mv ); 
                if ( index != -1 )
                        return (Vertex) vertexes.elementAt( index );
                else
                        return null;
        }
        
        private Vertex createVertex( int x, int y ) {
                Vertex mv = new Vertex( x, y );
                int index = vertexes.indexOf( mv );
                if ( index == -1 )
                        vertexes.addElement( mv );
                index = vertexes.indexOf( mv );
                
                return (Vertex) vertexes.elementAt( index );
        }

        private Line findLine( Vertex v1, Vertex v2 ) {
                Line line = new Line( v1, v2 );
                int index = lines.indexOf( line );
                if ( index != -1 )
                        return (Line) lines.elementAt( index );
                else
                        return null;
        }

        private Line createLine( Vertex v1, Vertex v2 ) {
                Line line = new Line( v1, v2 );
                int index = lines.indexOf( line );
                if ( index == -1 )
                        lines.addElement( line );
                index = lines.indexOf( line );
                line._map = this;

                return (Line) lines.elementAt( index );
        }

        public void flipLine( Line line ) {
                line.flipLine();

                notifyListeners( line.getBounds( 1 ) );
        }

        public void flipSectorLines( Sector s ) {
                for ( int i = 1; i < s.s_vertexes.length; ++i ) {
                        Line line = findLine( s.s_vertexes[i], s.s_vertexes[i-1] );
                        if ( line == null ) {
                                System.out.println( " error in flipSectorLines" );
                                continue;
                        }
                        line.flipLine();
                }

                Line line = findLine( s.s_vertexes[0], s.s_vertexes[ s.s_vertexes.length - 1 ] );
                line.flipLine();

                notifyListeners( s.getBounds( 1 ) );
        }

        public void flipVertexLines( Vertex v ) {
                
                Rectangle clip = new Rectangle( v.x, v.y, 0, 0 );
                
                for ( Iterator iter = lines.iterator(); iter.hasNext(); ) {
                        Line line = (Line) iter.next();
                        if ( line.hasVertRef( v ) ) {
                                line.flipLine();
                                
                                if ( line.verts[0].equals( v ) ) {
                                        clip.add( new Point( line.verts[1].x, line.verts[1].y ) );
                                } 
                                else {
                                        clip.add( new Point( line.verts[0].x, line.verts[0].y ) );
                                }
                        }
                             
                }
                notifyListeners( clip );
        }
        
        public void splitLine( Line line ) {
                Vertex result = 
                        new Vertex( ( line.verts[1].x - line.verts[0].x ) / 2 + line.verts[0].x,
                                    ( line.verts[1].y - line.verts[0].y ) / 2 + line.verts[0].y );
                Line[] newlines = new Line[] {
                        new Line( line.verts[0], result ),
                        new Line( result, line.verts[1] )
                                };
                
                
        }
        
        public LightSource findLight( Point3D p ) {
                for ( Iterator iter = lights.iterator(); iter.hasNext(); ) {
                        LightSource light = (LightSource) iter.next();
                        if ( light.pos.equals( p ) )
                                return light;
                }
                return null;
        }

        private void writeObject( ObjectOutputStream out )
                throws IOException 
        {
                out.defaultWriteObject();
        }
        
        private void readObject( ObjectInputStream in )
                throws IOException, ClassNotFoundException
        {
                in.defaultReadObject();
                listeners = new Vector();
                scriptListeners = new Vector();
                _selected = new Vector();
                ptsInProg = new int[2][100];
                _clip = new Rectangle( _mapRect );
                _transClip = null;
                numPtsInProg = 0;
                tranPts = new Point[] { 
                        new Point(),
                        new Point(),
                };

                drawTransient = false;
                _internalNotifyClip = new Rectangle();
                _defaultTexIndexes = new int[ NUM_DEF_TEXS ];
                
                for ( int i = 0; i < NUM_DEF_TEXS; ++i ) {
                        _defaultTexIndexes[i] = 
                                TextureManager.findTexture( _defaultTexs[i] );
                }
        }

        /**
         * save the map into a file named by filename, using serialization
         */
        public synchronized boolean saveAs( String filename ) {
                String fname = null;
                
                if ( filename == null )
                        fname = "Untitled." + JSFMAP_EXT;
                else
                        fname = filename;
                
                try {
                        FileOutputStream fos = new FileOutputStream( fname );
                        ObjectOutputStream oos = new ObjectOutputStream( fos );
                        oos.writeObject( this );
                        oos.close();
                        fos.close();
                        System.out.println( " Map save to " + fname + " successful." );
                        return true;
                }
                catch ( IOException exp ) {
                        exp.printStackTrace();
                }

                System.out.println( " Map save to " + fname + " failed." );
                return false;
                
        }
        
        /**
         * read the map from a file named by filename, using serialization
         */
        public static Map loadMap( String filename ) {
                
                try {
                        FileInputStream fis = new FileInputStream( filename );
                        ObjectInputStream ois = new ObjectInputStream( fis );
                        Object o = ois.readObject();
                        ois.close();
                        fis.close();
                        if ( o instanceof Map ) {
                                System.out.println( " Map open from " + filename + " successful." );
                                return (Map) o;
                        }
                }
                catch ( IOException exp ) {
                        exp.printStackTrace();
                }   
                catch ( ClassNotFoundException exp ) {
                        exp.printStackTrace();
                }
                
                System.out.println( " Map open from " + filename + " failed." );
                return null;
                
        }

        private synchronized String[] collapseTextures( int[] mapping ) {

                for ( int i = 0; i < mapping.length; ++i ) {
                        mapping[i] = -1;
                }

                int cur_num = 1;
                
                for ( Iterator iter = sectors.iterator(); iter.hasNext(); ) {
                        Sector s = (Sector) iter.next();

                        if ( s._floorTexIndex > 0 ) {
                                if ( mapping[ s._floorTexIndex ] == -1 ) {
                                        mapping[ s._floorTexIndex ] = cur_num++;
                                }
                        }

                        if ( s._ceilingTexIndex > 0 ) {
                                if ( mapping[ s._ceilingTexIndex ] == -1 ) {
                                        mapping[ s._ceilingTexIndex ] = cur_num++;
                                }
                        }
                }

                for ( Iterator iter = lines.iterator(); iter.hasNext(); ) {
                        Line line = (Line) iter.next();
                        
                        if ( line.line_sides[0] != null ) {
                                for ( int i = 0; i < 3; ++i ) {
                                        if ( mapping[ line.line_sides[0]._texIndexes[i] ] == -1 ) {
                                                mapping[ line.line_sides[0]._texIndexes[i] ] = cur_num++;
                                        }
                                }
                        }
                        
                        if ( line.line_sides[1] != null ) {
                                for ( int i = 0; i < 3; ++i ) {
                                        if ( mapping[ line.line_sides[1]._texIndexes[i] ] == -1 ) {
                                                mapping[ line.line_sides[1]._texIndexes[i] ] = cur_num++;
                                        }
                                }
                        }
                }

                String[] s = new String[ cur_num ];
                String[] texnames = TextureManager.getTexnames();
                
                for ( int i = 0; i < mapping.length; ++i ) {
                        if ( mapping[i] != -1 ) {
                                s[ mapping[ i ] ] = texnames[i];
                        }
                }

                return s;

        }
        
        private static byte[] _strbuf = new byte[1024];
        
        private String readString( DataInputStream dis ) throws IOException {
                int len = dis.read( _strbuf, 0, dis.readShort() );

                if ( len > 0 )
                        return new String( _strbuf, 0, len );
                else
                        return null;
        }

        private void writeString( DataOutputStream dos, String str ) throws IOException {
                if ( str == null )
                        dos.writeShort( 0 );
                else {
                        dos.writeShort( str.length() );
                        dos.writeBytes( str );
                }
        }
        
        /**
         * export the map into sdl-dg map format
         */
        public synchronized boolean exportMap( String filename ) {
                String fname = null;
                
                if ( filename == null )
                        fname = "Untitled." + DGMAP_EXT;
                else
                        fname = filename;
                
                try {
                        FileOutputStream fos = new FileOutputStream( fname );
                        DataOutputStream dos = new DataOutputStream( fos );

                        //
                        // Entity header (Entity.h)
                        //
                        
                        writeString( dos, name );
                        writeString( dos, name ); // alias
                        writeString( dos, name ); // icon

                        Vector sides    = new Vector();
                        sides.addElement( null );

                        for ( Iterator iter = lines.iterator(); iter.hasNext(); ) {
                                Line line = (Line) iter.next();
                                if ( line.line_sides[0] != null && ! sides.contains( line.line_sides[0] ) )
                                        sides.addElement( line.line_sides[0] );
                                if ( line.line_sides[1] != null && ! sides.contains( line.line_sides[1] ) )
                                        sides.addElement( line.line_sides[1] );
                        }
                         
                        dos.writeShort( vertexes.size() );
                        dos.writeShort( sides.size() );
                        dos.writeShort( lines.size() );
                        dos.writeShort( scripts.size() );
                        dos.writeShort( sectors.size() );
                        dos.writeShort( sectSpecials.size() );
                        
                        int[] mapping = new int[ TextureManager.getNumTexnames() ];
                        String[] actualTexnames = collapseTextures( mapping );
                        dos.writeShort( actualTexnames.length );
                        
                        System.out.println( " :: Exporting map with stats:\n" +
                                            "       verts  " + vertexes.size() + "\n" +
                                            "       sides  " + sides.size() + "\n" +
                                            "       lines  " + lines.size() + "\n" +
                                            "       sects  " + sectors.size() + "\n" +
                                            "       specs  " + sectSpecials.size() + "\n" +
                                            "     scripts  " + scripts.size() + "\n" +
                                            "       tex    " + actualTexnames.length + "\n" +
                                            "       lights " + lights.size() + "\n" );
                        
                        dos.writeShort( 0 ); // lightmaps
                        dos.writeByte( 0 ); // vis loaded flag
                        dos.writeShort( lights.size() ); // num_lights

                        //
                        // write the texture index
                        //
                        
                        for ( int i = 0; i < actualTexnames.length; ++i ) {
                                writeString( dos, actualTexnames[i] );
                        }

                        //
                        // write the verts
                        //
                        
                        for ( Iterator iter = vertexes.iterator(); iter.hasNext(); ) {
                                Vertex vert = (Vertex) iter.next();

                                dos.writeShort( vert.x );
                                dos.writeShort( vert.y );
                        }
                        
                        //
                        // write the sector specials
                        //
                        
                        for ( Iterator iter = sectSpecials.iterator(); iter.hasNext(); ) {
                                SectSpecial spec = (SectSpecial) iter.next();

                                
                                int flags = ( 1 << spec._type ) | ( spec._startExtended ? 0x0020 : 0 );
                                dos.writeShort( flags );
                                dos.writeShort( spec._travel );
                                dos.writeShort( spec._damage );
                                dos.writeShort( spec._sound );
                                

                                dos.writeByte( spec._speed );
                                dos.writeByte( spec._delay );
                                dos.writeByte( spec._tag );
                                dos.writeByte( spec._iterations );
                        }

                        //
                        // write the sectors
                        //

                        for ( Iterator iter = sectors.iterator(); iter.hasNext(); ) {
                                Sector s = (Sector) iter.next();

                                dos.writeShort( s.s_vertexes.length );
                                dos.writeShort( s.floorHeight );
                                dos.writeShort( s.ceilingHeight );
                                dos.writeShort( 0 ); // visible sectors
                                dos.writeShort( 0 ); // visible floors
                                dos.writeShort( 0 ); // visible ceilings
                                dos.writeShort( 0 ); // visible lines
                                dos.writeShort( mapping[ s._floorTexIndex ] );
                                dos.writeShort( mapping[ s._ceilingTexIndex ] );
                                dos.writeShort( -1 ); // floor lightmap
                                dos.writeShort( -1 ); // ceiling lightmap

                                int index =  sectSpecials.indexOf( s._special );
                                dos.writeShort( index );

                                for ( int i = 0; i < s.s_vertexes.length; ++i ) {
                                        short ind = (short) vertexes.indexOf( s.s_vertexes[i] );
                                        dos.writeShort( ind );
                                }
                        }

                        //
                        // write the sides
                        //
                        
                        for ( Iterator iter = sides.iterator(); iter.hasNext(); ) {
                                Side side = (Side) iter.next();
                                
                                // the first side, side 0 is null
                                if ( side == null ) {
                                        dos.writeShort( 0 );
                                        dos.writeShort( 0 );
                                        dos.writeShort( 0 );
                                        dos.writeShort( 0 );
                                        dos.writeShort( 0 );
                                        dos.writeShort( -1 );
                                        dos.writeShort( -1 );
                                        dos.writeShort( -1 );
                                        continue;
                                }

                                dos.writeShort( sectors.indexOf( side.sect ) );
                                dos.writeShort( lines.indexOf( side.line ) );
                                
                                //
                                // note: dg maps represent the sidedef textures in opposite order
                                // ( 0 == lower, 2 == upper )
                                //

                                for ( int i = 2; i >= 0; --i ) {
                                        dos.writeShort( mapping[ side._texIndexes[i] ] );
                                }

                                //
                                // lightmaps
                                //

                                dos.writeShort( -1 );
                                dos.writeShort( -1 );
                                dos.writeShort( -1 );
                        }
                                 
                        //
                        // write the line specials
                        //

                        for ( Iterator iter = scripts.iterator(); iter.hasNext(); ) {
                                Script script = (Script) iter.next();
                                script.export( dos );
                        }

                        //
                        // write the lines
                        //

                        for ( Iterator iter = lines.iterator(); iter.hasNext(); ) {
                                Line line = (Line) iter.next();
                                
                                dos.writeShort( vertexes.indexOf( line.verts[0] ) );
                                dos.writeShort( vertexes.indexOf( line.verts[1] ) );

                                dos.writeShort( sides.indexOf( line.line_sides[0] ) );
                                dos.writeShort( sides.indexOf( line.line_sides[1] ) );

                                dos.writeShort( line.trigger_flags );
                                dos.writeShort( line._script == null ? -1 : 
                                                scripts.indexOf( line._script ) );
                        }
                        

                        //
                        // write the lights
                        //

                        for ( Iterator iter = lights.iterator(); iter.hasNext(); ) {
                                LightSource light = (LightSource) iter.next();
                                
                                dos.writeShort( light.pos.x );
                                dos.writeShort( light.pos.y );
                                dos.writeShort( light.pos.z );

                                dos.writeByte( light.r );
                                dos.writeByte( light.g );
                                dos.writeByte( light.b );

                                dos.writeShort( light.intensity );
                        }


                        dos.close();
                        fos.close();
                        System.out.println( " Map export to " + fname + " successful." );
                        return true;
                }
                catch ( IOException exp ) {
                        exp.printStackTrace();
                }

                System.out.println( " Map export to " + fname + " failed." );
                return false;
                
                
        }


        public synchronized static Map importMap( String filename ) {

                Map map = new Map();
                
                try {
                        FileInputStream fis = new FileInputStream( filename );
                        DataInputStream dis = new DataInputStream( fis );

                        map.importMap( dis );

                        dis.close();
                        fis.close();
                        System.out.println( " Map import from " + filename + " successful." );

                }
                catch ( IOException exp ) {
                        exp.printStackTrace();
                        map = null;
                        System.out.println( " Map import from " + filename + " failed." );
;

                }
                
                return map;
        }


        private synchronized void importMap( DataInputStream dis ) throws IOException {
                // Entity header
                
                byte[] buf = new byte[1024];
                short len = 0;

                name = readString( dis );
                
                readString( dis ); // alias
                readString( dis ); // icon
                
                vertexes.setSize( dis.readShort() );
                Map.Side[] sides = new Map.Side[ dis.readShort() ];
                lines.setSize( dis.readShort() );
                scripts.setSize( dis.readShort() );
                sectors.setSize( dis.readShort() );
                sectSpecials.setSize( dis.readShort() );
                String[] texnames = new String[ dis.readShort() ];

                short num_lightmaps = dis.readShort();
                byte  vis_loaded    = dis.readByte();
                lights.setSize( dis.readShort() );
  
                   
                System.out.println( " :: Importing map with stats:\n" +
                                    "       verts  " + vertexes.size() + "\n" +
                                    "       sides  " + sides.length + "\n" +
                                    "       lines  " + lines.size() + "\n" +
                                    "       sects  " + sectors.size() + "\n" +
                                    "       specs  " + sectSpecials.size() + "\n" +
                                    "     scripts  " + scripts.size() + "\n" +
                                    "       tex    " + texnames.length + "\n" +
                                    "       lights " + lights.size() + "\n" );
                

                int[] mapping = new int[ texnames.length ];

                for ( int i = 0; i < texnames.length; ++i ) {
                        texnames[i] = readString( dis );
                        mapping[i] = TextureManager.findTexture( texnames[i] );
                }

                Point maxpt = new Point();
                
                for ( int i = 0; i < vertexes.size(); ++i ) {
                        Vertex vert = new Vertex( dis.readShort(),
                                                  dis.readShort() );
                        vertexes.set( i, vert );

                        // keep track of the range of vertices
                        if ( - vert.x > maxpt.x )
                                maxpt.x = - vert.x;
                        if ( vert.x > maxpt.x )
                                maxpt.x = vert.x;

                        if ( -vert.y > maxpt.y )
                                maxpt.y = - vert.y;
                        else if ( vert.y > maxpt.y )
                                maxpt.y = vert.y;
                                        
                }

                maxpt.x = ( maxpt.x / 1024 + 1 ) * 1024;
                maxpt.y = ( maxpt.y / 1024 + 1 ) * 1024;
                
                _mapSize.setSize( maxpt.x * 2, maxpt.y * 2 );
                _mapRect.setBounds( - maxpt.x, - maxpt.y,
                                    _mapSize.width, _mapSize.height );


                for ( int i = 0; i < sectSpecials.size(); ++i ) {
                        SectSpecial spec = new SectSpecial();
                        
                        int flags = dis.readShort();

                        spec._type = ( flags & 0x0001 ) == 0 ? spec.SECT_LIFT : spec.SECT_DOOR;
                        spec._startExtended = ( flags & 0x0020 ) != 0;
                        
                        spec._travel     = dis.readShort();
                        spec._damage     = dis.readShort();
                        spec._sound      = dis.readShort();
                        spec._speed      = dis.readByte();
                        spec._delay      = dis.readByte();
                        spec._tag        = dis.readByte();
                        spec._iterations = dis.readByte();

                        sectSpecials.set( i, spec );
                }

                for ( int i = 0; i < sectors.size(); ++i ) {

                        Sector s = new Sector();

                        s.s_vertexes = new Vertex[ dis.readShort() ];

                        s.floorHeight = dis.readShort();
                        s.ceilingHeight = dis.readShort();
                        
                        dis.readShort(); // visible sectors
                        dis.readShort(); // visible floors
                        dis.readShort(); // visible ceilings
                        dis.readShort(); // visible lines

                        s.setTex( s.FLOOR_TEX, mapping[ dis.readShort() ] );
                        s.setTex( s.CEILING_TEX, mapping[ dis.readShort() ] );
                        
                        dis.readShort(); // floor lightmap
                        dis.readShort(); // ceiling lightmap

                        int sind = dis.readShort();
                        
                        if ( sind >= 0 )
                                s._special = (SectSpecial) sectSpecials.elementAt( sind );
                        else
                                s._special = null;

                        for ( int j = 0; j < s.s_vertexes.length; ++j ) {
                                s.s_vertexes[j] = (Vertex) vertexes.elementAt( dis.readShort() );
                        }

                        sectors.set( i, s );
                }

                //
                // create the dummy lines
                //
                
                for ( int i = 0; i < lines.size(); ++i ) 
                        lines.set( i, new Line() );

                // read the dummy null side
                
                if ( dis.skip( 16 ) != 16 ) {
                        System.out.println( "ERROR SKIPPING 16 byte DUMMY SIDE!!" );
                        throw new IOException( "fuqqqt up" );
                }
                
                sides[0] = null;
                
                for ( int i = 1; i < sides.length; ++i ) {
                        Side side = new Side();

                        side.sect = (Sector) sectors.elementAt( dis.readShort() );
                        side.line = (Line) lines.elementAt( dis.readShort() );
                        
                        for ( int j = 2; j >= 0; --j ) {
                                side.setTexture( j, mapping[ dis.readShort() ] );
                        }

                        // lightmaps

                        dis.readShort();
                        dis.readShort();
                        dis.readShort();

                        sides[i] = side;
                }

                for ( int i = 0; i < scripts.size(); ++i ) {
                        scripts.set( i, new Script( dis ) );
                }
                
                for ( int i = 0; i < lines.size(); ++i ) {
                        Line line = (Line) lines.elementAt( i );

                        line.verts[0] = (Vertex) vertexes.elementAt( dis.readShort() );
                        line.verts[1] = (Vertex) vertexes.elementAt( dis.readShort() );
                        
                        line.line_sides[0] = (Side) sides[ dis.readShort() ];
                        line.line_sides[1] = (Side) sides[ dis.readShort() ];
                        
                        line.trigger_flags = dis.readShort();
                        int scr = dis.readShort();
                        if ( scr >= 0 )
                                line._script = (Script) scripts.elementAt( scr );
                        else
                                line._script = null;
                }

                for ( int i = 0; i < lights.size(); ++i ) {
                        LightSource light = new LightSource();

                        light.pos.x = dis.readShort();
                        light.pos.y = dis.readShort();
                        light.pos.z = dis.readShort();
                        
                        if ( ( light.r = dis.readByte() ) < 0 ) {
                                light.r = (short) ( 256 + light.r );
                        }
                        if ( ( light.g = dis.readByte() ) < 0 )
                                light.g = (short) ( 256 + light.g );
                        
                        if ( ( light.b = dis.readByte() ) < 0 )
                                light.b = (short) ( 256 + light.b );
                        
                        light.intensity = dis.readShort();
                        
                        lights.set( i, light );
                }
                
                System.out.println( " Map import completed." );

        }
                        

        //////////////////////////////////////////////////////////////////////////////
        //
        //
        // inner static classes and interfaces
        //
        //
        //////////////////////////////////////////////////////////////////////////////


        public static interface Listener {
                public void mapChanged( Rectangle clip );
                public void mapTransientChanged();
                public void mapSizeChanged();
        }

        public static interface ScriptListener {
                public void scriptsChanged();
        }

        //
        //
        //
        //
        //
        
        /**
         * base class for map elements
         */

        abstract public static class Element implements Serializable {

                public static final long serialVersionUID = 0;
                transient boolean isSelected = false;

                public Rectangle getBounds( int b ) {
                        Rectangle r = getSubclassBounds();
                        r.grow( b, b );
                        return r;
                }
                
                abstract public Point3D getCenterPoint3D();

                public Point getCenterPoint() {
                        return getCenterPoint3D().getPoint();
                }
                
                abstract public Rectangle getSubclassBounds();
                
        }

        /**
         * An interface for classes which want to modify a set of elements
         */

        public static interface ElementAction {
                public void doAction( Element elem );
        }

        /**
         * vertex
         */

        public static final class Vertex extends Element {
                
                int x, y;
                
                public Vertex( int x, int y ) {
                        this.x = x;
                        this.y = y;
                }
                
                public Point3D getCenterPoint3D() {
                        return new Point3D( x, y, 0 );
                }

                public boolean equals( Object o ) {
                        if ( o instanceof Vertex ) {
                                return ( ( (Vertex) o ).x == x && ( (Vertex) o ).y == y );
                        }
                        return super.equals( o );
                }

                /**
                 * Vertex subclass bounds is a 2x2 rectangle
                 */
                public Rectangle getSubclassBounds() {
                        Rectangle r = new Rectangle( x-1, y-1, 2, 2 );
                        return r;
                }

        }

        /**
         * line
         */

        public static final class Line extends Element {
                static final long serialVersionUID=-139135798021765249L;

                static final short REPEATABLE = 0x0001;
                static final short FRONT_OK   = 0x0002;
                static final short BACK_OK    = 0x0004;
                static final short PLAYER_OK  = 0x0008;
                static final short MONSTER_OK = 0x0010;
                static final short ITEM_OK    = 0x0020;
                
                static final String[] trigger_flagnames = new String[] {
                        "Repeatable",
                        "Front Ok",
                        "Back Ok",
                        "Player Ok",
                        "Monster Ok",
                        "Item Ok"
                };

                Vertex[] verts = new Vertex[2];
                
                Side[]   line_sides     = new Side[2];
                
                Map _map = null;

                Script _script = null;
                short trigger_flags   = 0x003f;

                Line() {}

                public Line( Vertex v1, Vertex v2 ) {
                        verts[0] = v1;
                        verts[1] = v2;
                }

                public Point3D getCenterPoint3D() {
                        return new Point3D( ( verts[0].x + verts[1].x ) / 2,
                                            ( verts[0].y + verts[1].y ) / 2,
                                            0 );
                }
                
                public boolean equals( Object o ) {
                        if ( o instanceof Line ) {
                                Line line = (Line) o;
                                return ( ( line.verts[0].equals( verts[0] ) && 
                                           line.verts[1].equals( verts[1] ) ) ||
                                         ( line.verts[0].equals( verts[1] ) && 
                                           line.verts[1].equals( verts[0] ) ) );
                        }        
                        return super.equals( o );
                }
                
                public Rectangle getSubclassBounds() {
                        Rectangle r = new Rectangle( Math.min( verts[0].x, verts[1].x ) - 4,
                                                     Math.min( verts[0].y, verts[1].y ) - 4,
                                                     Math.abs( verts[0].x - verts[1].x ) + 8,
                                                     Math.abs( verts[0].y - verts[1].y ) + 8 );
                        return r;
                }
                
                int swapVertexRefs( Vertex old, Vertex target ) {
                        
                        int num = 0;
                        
                        if ( verts[0] == old ) {
                                verts[0] = target;
                                ++num;
                        }

                        if ( verts[1] == old ) {
                                verts[1] = target;
                                ++num;
                        }
                        return num;
                }
                
                private void flipLine() {
                        Vertex tmp = verts[0];
                        verts[0] = verts[1];
                        verts[1] = tmp;

                        Side s_tmp = line_sides[0];
                        line_sides[0] = line_sides[1];
                        line_sides[1] = s_tmp;

                }


                private boolean hasVertRef( Vertex v ) {
                        return ( verts[0] == v || verts[1] == v );
                }

                private boolean hasVert( Vertex v ) {
                        return ( hasVertRef( v ) || verts[0].equals( v ) || verts[1].equals( v ) );
                }

                public Script setScript( int num ) {
                        
                        Script script = null;
                        for ( Iterator iter = _map.scripts.iterator(); iter.hasNext(); ) {
                                script = (Script) iter.next();
                                if ( script.num == num ) 
                                        break;
                                script = null;
                        }
                        
                        if ( script == null ) {
                                script = new Script( num );
                                _map.scripts.addElement( script );
                                _map.notifyScriptListeners();
                        }
                        
                        _script = script;
                        
                        _map.notifyListeners( getBounds( 0 ) );

                        return script;

                }
        }
        

        /**
         * side
         */

        public static final class Side extends Element implements Serializable {

                public static final long serialVersionUID = 0;

                Sector sect = null;
                Line   line = null;
                
                public static final int UPPER_INDEX = 0;
                public static final int MAIN_INDEX  = 1;
                public static final int LOWER_INDEX = 2;

                private String[] _textures          = new String[3];
                transient private int[] _texIndexes = new int[3];

                public Side() {
                }

                public Side( Line line, Sector sect, int[] texs ) {
                        this.line = line;
                        this.sect = sect;

                        for ( int i = 0; i < 3; ++i ) {
                                setTexture( i, texs[i] );
                        }
                }
                
                public Side getOpposite() {
                        return ( this == line.line_sides[0] ) ? line.line_sides[1] : line.line_sides[0];
                }
                
                public boolean hasSection( int which, Side other ) {

                        if ( which == MAIN_INDEX ) {
                                return ( other == null );
                        }

                        else if ( which == LOWER_INDEX ) {
                                return ( other != null && 
                                         ( sect.floorHeight < other.sect.floorHeight || other.sect.isLift() ) );
                        }

                        else {
                                return ( other != null && 
                                         ( sect.ceilingHeight > other.sect.ceilingHeight || other.sect.isDoor() ) );
                        }
                }
                
                public Point3D getCenterPoint3D() {
                        Point3D p = line.getCenterPoint3D();
                        p.z = ( sect.floorHeight + sect.ceilingHeight ) / 2;
                        return p;
                }
                
                public Rectangle getSubclassBounds() {
                        return line.getSubclassBounds();
                }

                void setTexture( int which, int texindex ) {
                        _texIndexes[ which ] = texindex;
                        _textures[ which ] = TextureManager.getTexname( texindex );
                }
                
                int getTexture( int which ) {
                        return _texIndexes[ which ];
                }
                
                private void readObject( ObjectInputStream in )
                        throws IOException, ClassNotFoundException
                {
                        in.defaultReadObject();
                        
                        _texIndexes = new int[3];

                        for ( int i = 0; i < 3; ++i ) {
                                _texIndexes[i] = TextureManager.findTexture( _textures[i] );
                        }
                }
        }

        public static final class SectSpecial implements Serializable {

                public static final long serialVersionUID = 0;
                
                public static final short SECT_DOOR = 0;
                public static final short SECT_LIFT = 1;
                
                public static final String[] names = new String[] {
                        "Door",
                        "Lift"
                };
                
                public static final int SND_SILENT        = 0;
                public static final int SND_METAL_RATCHET = 1;
                public static final int SND_LIGHT_STONE   = 2;
                public static final int SND_HEAVY_STONE   = 3;
                public static final int SND_SCRAPE_METAL  = 4;
                public static final int SND_CREAK_WOOD    = 5;

                public static final String[] sndNames = new String[] {
                        "Silent",
                        "Metal Ratchet",
                        "Light Stone",
                        "Heavy Stone",
                        "Scrape Metal",
                        "Creak Wood"
                };

                byte    _tag = 0;
                short   _type = 0;
                short   _travel = 0;
                byte    _speed = 0;
                byte    _delay = 0;
                short   _damage = 0;
                short   _sound = 0;    
                boolean _startExtended = false;
                byte    _iterations = 0;

        }
        
        /**
         * sector
         */ 

        public static final class Sector extends Element implements Serializable {
                
                public static final long serialVersionUID = 0;

                public static final int CEILING_TEX = 0;
                public static final int FLOOR_TEX   = 1;
                
                int num = 0;
                Map _map;
                
                SectSpecial _special = null;
                
                int floorHeight = 0;
                int ceilingHeight  = 256;
                short brightness = 255;
                
                Vertex[] s_vertexes;
                
                private String _floorTex = null;
                private String _ceilingTex  = null;
                
                transient private int _floorTexIndex = 0;
                transient private int _ceilingTexIndex = 0;
                
                public boolean isDoor() { return _special != null && 
                                                  _special._type == SectSpecial.SECT_DOOR; }

                public boolean isLift() { return _special != null && 
                                                  _special._type == SectSpecial.SECT_LIFT; }

                public SectSpecial createNewSpecial() {

                        if ( _special != null )
                                return _special;
                        
                        byte max_spec = 0;
                        
                        for ( Iterator iter = _map.sectSpecials.iterator(); iter.hasNext(); ) {
                                SectSpecial spec = (SectSpecial) iter.next();
                                max_spec = (byte) Math.max( max_spec, spec._tag );
                        }
                        
                        ++max_spec;
                        
                        SectSpecial spec = new SectSpecial();
                        _map.sectSpecials.addElement( spec );

                        spec._tag = max_spec;
                        
                        _special = spec;
                        
                        _map.notifyListeners( getBounds( 0 ) );
                        return spec;
                }

                public void deleteSpecial() {
                        if ( _special == null )
                                return;
                        
                        _map.sectSpecials.removeElement( _special );
                        _special = null;
                        _map.notifyListeners( getBounds( 0 ) );
                }

                Sector() {}

                public Sector( int[][] points, int numPts ) {
                        
                        s_vertexes = new Vertex[numPts];
                        
                        //
                        // make sure the sector is counter-clockwise wound
                        //
                        
                        double ang = 0;
                        
                        double wind = 0;
                        for ( int i = 2; i < numPts; ++i ) {
                                ang = VMath.crossProductMagPolyline( points[0][i-2], points[1][i-2],
                                                                     points[0][i-1], points[1][i-1],
                                                                     points[0][i],   points[1][i] );

                                wind += ang;
                        }

                        ang = VMath.crossProductMagPolyline( points[0][numPts-2], points[1][numPts-2],
                                           points[0][numPts-1], points[1][numPts-1],
                                           points[0][0],     points[1][0] );
                        
                        wind += ang;
                        
                        ang = VMath.crossProductMagPolyline( points[0][numPts-1], points[1][numPts-1],
                                           points[0][0],     points[1][0],
                                           points[0][1],     points[1][1] );

                        wind += ang;
                        
                        //
                        // negative wound is correct counter-clockwise (cross product of rotation, 
                        // negative z cross product)
                        //

                        if ( wind < 0 ) {
                                for ( int i = 0; i < numPts; ++i ) {
                                        Vertex v = new Vertex( points[0][i], points[1][i] );
                                        s_vertexes[i] = v;
                                }
                        }

                        //
                        // reverse the winding
                        //
                        
                        else {
                                for ( int i = 0; i < numPts; ++i ) {
                                        Vertex v = new Vertex( points[0][i], points[1][i] );
                                        s_vertexes[numPts-i-1] = v;    
                                }
                        }
                }

                public Point3D getCenterPoint3D() {
                        Point3D p = new Point3D();
                        
                        Rectangle rect = getPolygon().getBounds();

                        p.x = rect.x + ( rect.width / 2 );
                        p.y = rect.y + ( rect.height / 2 );

                        p.z = ( floorHeight + ceilingHeight ) / 2;

                        return p;
                }
                
                public Rectangle getSubclassBounds() {
                        return getPolygon().getBounds();
                }
                
                int getTexIndex( int type ) {
                        if ( type == CEILING_TEX ) {
                                return _ceilingTexIndex;
                        }
                        return _floorTexIndex;
                }

                void setTex( int type, int index ) {
                        if ( type == CEILING_TEX ) {
                                _ceilingTexIndex = index;
                                _ceilingTex = TextureManager.getTexname( index );
                        }
                        else {
                                _floorTexIndex = index;
                                _floorTex = TextureManager.getTexname( index );
                        }
                }
                
                private void readObject( ObjectInputStream in )
                        throws IOException, ClassNotFoundException
                {
                        in.defaultReadObject();
                        
                        _floorTexIndex = TextureManager.findTexture( _floorTex );
                        _ceilingTexIndex = TextureManager.findTexture( _ceilingTex );
                }
                
                Polygon getPolygon() {
                        Polygon poly = new Polygon();
                        
                        for ( int i = 0; i < s_vertexes.length; ++i ) {
                                poly.addPoint( s_vertexes[i].x, s_vertexes[i].y );
                        }
                        return poly;
                }

                private boolean removeVertex( Vertex v ) {
                        boolean found = false;

                        for ( int i = 0; i < s_vertexes.length; ++i ) {
                                if ( s_vertexes[i] == v ) 
                                        found = true;
                        }
                        
                        if ( !found )
                                return false;
                        
                        if ( s_vertexes.length > 1 ) {
                                Vertex[] remain = new Vertex[ s_vertexes.length - 1 ];

                                int n = 0;

                                for ( int i = 0; i < s_vertexes.length; ++i ) {
                                        if ( s_vertexes[i] != v )
                                                remain[ n++ ] = s_vertexes[i];
                                }
                                
                                s_vertexes = remain;
                        }

                        return true;
                }
                
                private void swapVertexRefs( Vertex old, Vertex target ) {
                        boolean found = false;
                        
                        for ( int i = 0; i < s_vertexes.length; ++i ) {

                                //
                                // swap all references, only
                                //

                                if ( s_vertexes[i] == old ) {
                                        s_vertexes[i] = !found ? target : null;
                                        found = true;
                                }

                                //
                                // if they are equal anyway, we have a duplicate
                                // that needs to be removed
                                //

                                else if ( s_vertexes[i] != target && 
                                          s_vertexes[i].equals( old ) ) {
                                        s_vertexes[i] = null;
                                }
                        }
                        
                        removeVertex( null );
                }
                
                private boolean hasVertex( Vertex v ) {
                        for ( int i = 0; i < s_vertexes.length; ++i ) {
                                if ( s_vertexes[i] == v ) {
                                        return true;
                                }
                        }
                        return false;
                }

                public int getFloorHeight() {
                        return floorHeight;
                }
                public int getCeilingHeight() {
                        return ceilingHeight;
                }
                        
                public void setFloorHeight( int h, boolean cache ) {
                        floorHeight = h;

                        if ( cache )
                                _map.addCacheClip( getBounds( 0 ) );
                        else
                                _map.notifyListeners( getBounds( 0 ) );
                }
                public void setCeilingHeight( int h, boolean cache ) {
                        ceilingHeight = h;

                        if ( cache )
                                _map.addCacheClip( getBounds( 0 ) );
                        else
                                _map.notifyListeners( getBounds( 0 ) );
                }
        }

        public static final class LightSource extends Element implements Serializable {
                
                public static final long serialVersionUID = -8104542567637257259l;

                Point3D pos = new Point3D();
                short r = 255, g = 255, b = 255;
                int intensity;
                
                /**
                 * LightSource subclass bounds is a 16x16 rectangle
                 */
                public Rectangle getSubclassBounds() {
                        return new Rectangle( pos.x - 8, pos.y - 8, 16, 16 );
                }

                public Point3D getCenterPoint3D() {
                        return new Point3D( pos.x, pos.y, pos.z );
                }
                
                public void setColor( int ar, int ag, int ab ) {
                        r = (short) ar;
                        g = (short) ag;
                        b = (short) ab;
                }
                
                public void setIntensity( int i ) {
                        intensity = i;
                }
                
        }

        public static final class Entity extends Element implements Serializable {
                
                Point3D pos;
                //                int x, y, z;
                
                /**
                 * LightSource subclass bounds is a 16x16 rectangle
                 */
                public Rectangle getSubclassBounds() {
                        return new Rectangle( pos.x - 8, pos.y - 8, 16, 16 );
                }
                
                public Point3D getCenterPoint3D() {
                        return new Point3D( pos.x, pos.y, pos.z );
                }
        }

        public static void main( String[] args ) throws Exception {
                Map map = Map.importMap( args[0] );
                map.exportMap( args[0] + "_tst" );
        }
}





