import javax.swing.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

public class MultiSlider extends JComponent {
                        
        JSlider[]   _sliders = null;
        JLabel[]     _labels = null;
        JTextField[] _fields = null;
             
        Vector _listeners = new Vector();

        public MultiSlider( int num_rows,
                            String[] labels,
                            int[] minvals,
                            int[] maxvals,
                            boolean[] showVals,
                            boolean[] editVals ) {
                                
                _sliders = new JSlider[ num_rows ];
                _labels  = new JLabel[ num_rows ];
                _fields  = new JTextField[ num_rows ];
                
                setLayout( new GridBagLayout() );
                GridBagConstraints gbc = new GridBagConstraints();

                gbc.gridx = gbc.gridy = 0;
                gbc.weightx = gbc.weighty = 0;
                gbc.fill = gbc.HORIZONTAL;
                gbc.insets = new Insets( 2, 2, 2, 2 );
                
                for ( int i = 0; i < num_rows; ++i ) {

                        gbc.weightx = 0;
                        gbc.gridx = 0;
                                        
                        if ( labels[i] != null ) {
                                _labels[i] = new JLabel( labels[i] );
                                add( _labels[i], gbc );
                        }

                        ++gbc.gridx;
                                        
                        _sliders[i] = new JSlider( JSlider.HORIZONTAL,
                                                   minvals[i],
                                                   maxvals[i],
                                                   0 );
                        _sliders[i].addChangeListener( _slideListener );

                        gbc.weightx = 1;
                        add( _sliders[i], gbc );
                        gbc.weightx = 0;

                        ++gbc.gridx;
                        if ( showVals[i] ) {
                                _fields[i] = new JTextField( 3 );
                                _fields[i].addActionListener( _fieldListener );
                                _fields[i].addFocusListener( _fieldFocusListener );

                                _fields[i].setEditable( editVals[i] );
                                _fields[i].setMinimumSize( _fields[i].getPreferredSize() );
                                add( _fields[i], gbc );
                        }

                        ++gbc.gridy;
                }
        }

        private ChangeListener _slideListener = new ChangeListener() {
                        public void stateChanged( ChangeEvent evt ) {
                                Object source = evt.getSource();
                                
                                //                                System.out.println( " slide changed: " + source );
                                                
                                for ( int i = 0; i < _sliders.length; ++i ) {
                                        if ( source == _sliders[i] ) {
                                                _fields[i].setText( "" + _sliders[i].getValue() );
                                                
                                                notifyListeners( source, i, _sliders[i].getValue() );
                                                
                                                return;
                                        }
                                }
                        }
                };

        private ActionListener _fieldListener = new ActionListener() {
                        public void actionPerformed( ActionEvent evt ) {
                                setValueFromField( evt.getSource() );

                        }
                };
        
        private FocusListener _fieldFocusListener = new FocusAdapter() {
                        public void focusLost( FocusEvent evt ) {
                                setValueFromField( evt.getSource() );
                        }
                };

        public void addListener( Listener listener ) {
                _listeners.addElement( listener );
        }

        public void removeListener( Listener listener ) {
                _listeners.removeElement( listener );
        }

        private void notifyListeners( Object source, int which, int value ) {
                for ( Iterator iter = _listeners.iterator(); iter.hasNext(); ) {
                        Listener listener = (Listener) iter.next();
                        if ( listener == source )
                                continue;
                        
                        if ( listener.isListening() )
                                listener.changed( which, value );
                }
        }

        private void setValueFromField( Object source ) {
                for ( int i = 0; i < _sliders.length; ++i ) {
                        if ( source == _fields[i] ) {
                                int val;
                                                
                                try {
                                        val = Integer.parseInt( _fields[i].getText() );
                                        
                                        if ( _sliders[i].getValue() == val )
                                                return;
                                        
                                        _sliders[i].setValue( val );
                                        
                                        if ( _sliders[i].getValue() != val ) {
                                                _fields[i].setText( "" + _sliders[i].getValue() );
                                        }
                                }
                                catch ( NumberFormatException exp ) {
                                        Toolkit.getDefaultToolkit().beep();
                                }
                                                
                                return;
                        }
                }
        }
        
        public int getValueAt( int which ) {
                return _sliders[which].getValue();
        }

        public void setValueAt( int which, int value ) {
                _sliders[which].setValue( value );
        }

        abstract public static class Listener {
                private boolean _isListening = true;
                abstract public void changed( int which, int value );
                public boolean isListening() {
                        return _isListening;
                }
                protected void setListening( boolean val ) {
                        _isListening = val;
                }
        }
}
                
