/*
 * VMath.java:
 *
 * Copyright (C) 2000 John Watson, jwatson@tempusmud.com
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */
import java.util.*;
import java.lang.*;


public class VMath {

        public static double square( double x ) {
                return x * x;
        }

        public static double length( double x, double y ) {
                return Math.sqrt( square( x ) + square( y ) );
        }
        
        public static boolean inRange( double x, double y, double ax, double ay, double range ) {
                if ( Math.abs( x - ax ) > range ||
                     Math.abs( y - ay ) > range )
                        return false;
                return ( distance( x, y, ax, ay ) <= range );
        }
        
        public static double distance( double x, double y, double ax, double ay ) {
                return Math.sqrt( square( x - ax ) + square( y - ay ) );
        }

        public static int[] normal( int x, int y ) {
                return new int[] { y, -x };
                
        }

        public static void unitNormal( int x, int y, double[] res ) {
                double len = length( x, y );
                res[0] =   (double) y / len;
                res[1] = - (double) x / len;
        }

        public static void unitVector( int x, int y, double[] res ) {
                double len = length( x, y );
                res[0] = (double) x / len;
                res[1] = (double) y / len;
        }

        public static double crossProductMag( int ax, int ay, int bx, int by ) {
                return ( ax * by ) - ( ay * bx );
        }

        public static double crossProductMagPolyline( int ax, int ay, int bx, int by, int cx, int cy ) {
                return crossProductMag( bx - ax, by - ay, cx - bx, cy - by );
        }
        
        public static double angle( int ax, int ay, int bx, int by, int cx, int cy ) {
                return ( angle( bx - ax, by - ay, cx - bx , cy - by ) );
        }

        public static double angle( int ax, int ay, int bx, int by ) {
                System.out.println( "     angle : " + ax + ", " + ay + ", " + bx + ", " + by );
                return Math.toDegrees( Math.atan2( by, bx ) - Math.atan2( ay, ax ) );
                /*
                double cross = crossProductMag( ax, ay, bx, by );
                double[] lens = new double[] {
                        length( ax, ay ),
                        length( bx, by )
                };
                
                double sin_theta = cross / ( lens[0] * lens[1] );

                return Math.toDegrees( Math.asin( sin_theta ) );
                */
        }
        
        /**
         * returns Double.POSITIVE_INFINITY if closest point to line is not contained
         * in line segment
         */
        public static double distFromLine( int x, int y, 
                                           int ax, int ay, 
                                           int bx, int by, 
                                           boolean segment ) {
                
                if ( ax != bx || ay != by ) {
                        double denom = square( bx - ax ) + square( by - ay );
                        
                        double u = ( x - ax ) * ( bx - ax ) + ( y - ay ) * ( by - ay );
                        u /= denom;

                        if ( segment ) {

                                if ( u < 0 || u > 1 ) {
                                        return Double.POSITIVE_INFINITY;
                                }
                        }
                        
                        double[] test = new double[] {
                                ax + u * ( bx - ax ),
                                ay + u * ( by - ay )
                        };

                        return distance( test[0], test[1], x, y );
                }
                else {
                        return distance( x, y, ax, ay);
                }
        }
}

