/*
 * ViewControl.java:
 *
 * Copyright (C) 2000 John Watson, jwatson@tempusmud.com
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

public class ViewControl implements Map.Listener {

        public static int MODE_NONE       = 0;
        public static int MODE_SELECT     = 1;
        public static int MODE_NEW_SECTOR = 2;
        public static int MODE_NEW_LIGHT  = 3;
        public static int MODE_NEW_ENTITY = 4;
        public static int MODE_SET_CLIP   = 5;
        
        private JMenu _menu;
        private Map _map;

        private int _gridSize = 64;

        private boolean _makingNewSector = false;

        private Vector _vcPairs = new Vector( 3 );

        private JRadioButtonMenuItem _selectItem    = 
                new JRadioButtonMenuItem( "Select Mode (s)", true );
        private JRadioButtonMenuItem _newSectorItem = 
                new JRadioButtonMenuItem( "New Sector Mode (a)", false );
        private JRadioButtonMenuItem _newLightItem  = 
                new JRadioButtonMenuItem( "New Light Mode (g)", false );
        private JRadioButtonMenuItem _newEntityItem  = 
                new JRadioButtonMenuItem( "New Entity Mode (e)", false );
        private JRadioButtonMenuItem _clipItem  = 
                new JRadioButtonMenuItem( "Clip Mode (c)", false );
        
        private JCheckBoxMenuItem _snapItem = new JCheckBoxMenuItem( "Snap to Grid (c-g)", true );

        private JMenuItem _gridLargerItem   = new JMenuItem( "Grid Larger ([)", 'G' );
        private JMenuItem _gridSmallerItem  = new JMenuItem( "Grid Smaller (])", 'S' );
        
        private JMenuItem _clearItem      = new JMenuItem( "Clear (c-k)", 'L' );

        private JMenuItem _mergeVertsItem = new JMenuItem( "Merge Verts (c-m)", 'M' );
        private JMenuItem _flipLineItem   = new JMenuItem( "Flip Line(s) (c-f)", 'F' );
        private JMenuItem _splitLineItem  = new JMenuItem( "Split Line (c-i)", 'D' );

        private JMenuItem _copy  = new JMenuItem( "Copy", 'C' );
        private JMenuItem _cut   = new JMenuItem( "Cut", 'u' );
        private JMenuItem _paste = new JMenuItem( "Paste", 'P' );

        private JMenuItem _stairsItem = new JMenuItem( "Create Stairs (c-t)", 'T' );
        
        private JMenuItem _shrinkClipItem = new JMenuItem( "Shrinkwrap Clip (c-w)", 'w' );
        private JMenuItem _expandClipItem = new JMenuItem( "Expand Clip (c-a)", 'x' );

        //
        // selection menu stuff
        //

        private JMenu _selectMenu = null;
        private JMenuItem _selectAllItem       = new JMenuItem( "Select All" );
        private JMenuItem _unselectItem        = new JMenuItem( "Unselect" );
        private JMenuItem _invertSelectionItem = new JMenuItem( "Invert Selection" );

        public static int AUTO_NONE     = 0;
        public static int AUTO_SECTORS  = 1;
        public static int AUTO_LINES    = 2;
        public static int AUTO_VERTS    = 3;
        public static int AUTO_LIGHTS   = 4;
        public static int AUTO_ENTITIES = 5;

        private JRadioButtonMenuItem _autoSelectSectsItem    = 
                new JRadioButtonMenuItem( "Autoselect Sects", true );
        private JRadioButtonMenuItem _autoSelectLinesItem    = 
                new JRadioButtonMenuItem( "Autoselect Lines", false );
        private JRadioButtonMenuItem _autoSelectVertsItem    = 
                new JRadioButtonMenuItem( "Autoselect Verts", false );
        private JRadioButtonMenuItem _autoSelectLightsItem   = 
                new JRadioButtonMenuItem( "Autoselect Lights", false );
        private JRadioButtonMenuItem _autoSelectEntitiesItem = 
                new JRadioButtonMenuItem( "Autoselect Entities", false );


        public ViewControl( JMenu menu, JMenu selectMenu ) {

                _menu = menu;
                _selectMenu = selectMenu;
                
                _infoBar._modeField.setText( ( _selectItem.isSelected() ? "Select" :
                                               ( _newLightItem.isSelected() ? "New Light" :
                                                 ( _clipItem.isSelected() ? "Set Clip" : "New Sector" ) ) ) );
                
                _infoBar._gridSizeField.setText( "" + _gridSize + 
                                                 ( _snapItem.isSelected() ? " (S)" : "" ) );
                
                _selectItem.addActionListener(      menuListener );
                _newSectorItem.addActionListener(   menuListener );
                _newLightItem.addActionListener(    menuListener );
                _clipItem.addActionListener(        menuListener );
                _clearItem.addActionListener(       menuListener );
                _gridSmallerItem.addActionListener( menuListener );
                _gridLargerItem.addActionListener(  menuListener );
                _mergeVertsItem.addActionListener(  menuListener );
                _flipLineItem.addActionListener(    menuListener );
                _splitLineItem.addActionListener(   menuListener );
                _stairsItem.addActionListener(      menuListener );
                _shrinkClipItem.addActionListener(  menuListener );
                _shrinkClipItem.addActionListener(  menuListener );

                _snapItem.addActionListener(             menuListener );
                
                 _selectAllItem.addActionListener(       menuListener );
                 _unselectItem.addActionListener(        menuListener );
                 _invertSelectionItem.addActionListener( menuListener );

                _copy.addActionListener(  clipboardListener );
                _cut.addActionListener(   clipboardListener );
                _paste.addActionListener( clipboardListener );
                
                _copy.setEnabled(           false );
                _cut.setEnabled(            false );
                _paste.setEnabled(          false );
                _mergeVertsItem.setEnabled( false );
                _flipLineItem.setEnabled(   false );
                _splitLineItem.setEnabled(  false );
                _stairsItem.setEnabled(     false );
                
                ButtonGroup bg = new ButtonGroup();
                bg.add( _selectItem );
                bg.add( _newSectorItem );
                bg.add( _newLightItem );
                bg.add( _clipItem );

                _menu.add( _selectItem );
                _menu.add( _newSectorItem );
                _menu.add( _newLightItem );
                _menu.add( _clipItem );

                _menu.addSeparator();
                
                _menu.add( _snapItem );
                _menu.add( _gridLargerItem );
                _menu.add( _gridSmallerItem );
                _menu.addSeparator();
                
                _menu.add( _copy );
                _menu.add( _cut );
                _menu.add( _paste );                
                _menu.add( _clearItem );

                _menu.addSeparator();

                _menu.add( _mergeVertsItem );
                _menu.add( _flipLineItem );
                _menu.add( _splitLineItem );
                _menu.add( _stairsItem );
                
                _menu.addSeparator();
                
                _menu.add( _shrinkClipItem );
                _menu.add( _expandClipItem );

                //
                // setup the select menu
                //

                _selectMenu.add( _selectAllItem );
                _selectMenu.add( _unselectItem );
                _selectMenu.add( _invertSelectionItem );
                _selectMenu.addSeparator();

                bg = new ButtonGroup();
                bg.add( _autoSelectSectsItem );
                bg.add( _autoSelectLinesItem );
                bg.add( _autoSelectVertsItem );
                bg.add( _autoSelectLightsItem );
                bg.add( _autoSelectEntitiesItem );

                _selectMenu.add( _autoSelectSectsItem );
                _selectMenu.add( _autoSelectLinesItem );
                _selectMenu.add( _autoSelectVertsItem );
                _selectMenu.add( _autoSelectLightsItem );
                _selectMenu.add( _autoSelectEntitiesItem );
                
        }

        public void setMap( Map map ) {

                if ( _map != null )
                        _map.listeners.removeElement( this );
                
                _map = map;
                
                if ( _map != null ) {
                        _map.listeners.addElement( this );
                }
                
                performVCPairActions( new VCPairAction() {
                                public void doAction( VCPair vcp ) {
                                        vcp._controller.setMap( _map );
                                        vcp._view.setMap( _map );
                                }
                        } );
        }
        
        private VCPairAction _repaintAction = new VCPairAction() {
                        public void doAction( VCPair vcp ) {
                                vcp._view.repaint();
                                // TODO: remove this ugly hack
                                vcp._view.getRule( View2d.HORIZONTAL ).repaint();
                                vcp._view.getRule( View2d.VERTICAL ).repaint();
                        }
                };

        public void setGridSize( int size ) {
                Dimension mapSize = _map.getMapSize();
                
                if ( size >= 10 && 
                     size <= mapSize.width / 2 && 
                     size <= mapSize.height / 2 ) {
                        _gridSize = size;
                }
                
                _infoBar._gridSizeField.setText( "" + _gridSize + 
                                                 ( _snapItem.isSelected() ? " (S)" : "" ) );
                
                performVCPairActions( _repaintAction );
        }
        
        public int getGridSize() {
                return _gridSize;
        }

        public Point3D getSnapped( int x, int y, int z ) {
                int halfGrid = _gridSize / 2;
                Point3D p = new Point3D();
                if ( x > 0 )
                        p.x = ( ( x + halfGrid ) / _gridSize ) * _gridSize;
                else
                        p.x = ( ( x - halfGrid ) / _gridSize ) * _gridSize;
                        
                if ( y > 0 )
                        p.y = ( ( y + halfGrid ) / _gridSize ) * _gridSize;
                else
                        p.y = ( ( y - halfGrid ) / _gridSize ) * _gridSize;

                if ( z > 0 )
                        p.z = ( ( z + halfGrid ) / _gridSize ) * _gridSize;
                else
                        p.z = ( ( z - halfGrid ) / _gridSize ) * _gridSize;

                return p;
        }

        public Point3D getSnapped( Point3D p ) {
                return getSnapped( p.x, p.y, p.z );
        }
        
        public int getAutoSelectMode() {
                if ( _autoSelectSectsItem.isSelected() )
                        return AUTO_SECTORS;
                if ( _autoSelectLinesItem.isSelected() )
                        return AUTO_LINES;
                if ( _autoSelectVertsItem.isSelected() )
                        return AUTO_VERTS;
                if ( _autoSelectLightsItem.isSelected() )
                        return AUTO_LIGHTS;
                if ( _autoSelectEntitiesItem.isSelected() )
                        return AUTO_ENTITIES;
                
                return AUTO_NONE;
        }

        public boolean isAutoSelect( int mode ) {
                return ( getAutoSelectMode() == mode );
        }

        public int getMode() {
                if ( _selectItem.isSelected() )
                        return MODE_SELECT;
                else if ( _newSectorItem.isSelected() )
                        return MODE_NEW_SECTOR;
                else if ( _newLightItem.isSelected() )
                        return MODE_NEW_LIGHT;
                else if ( _newEntityItem.isSelected() )
                        return MODE_NEW_ENTITY;
                else if ( _clipItem.isSelected() )
                        return MODE_SET_CLIP;
                
                return MODE_NONE;
        }
        
        public boolean isMode( int mode ) {
                return ( getMode() == mode );
        }

        public boolean isSnapEnabled() {
                return _snapItem.isSelected();
        }


        private Map.ElementAction _flipAction = new Map.ElementAction() {
                        public void doAction( Map.Element elem ) {
                                if ( elem instanceof Map.Line ) {
                                        _map.flipLine( (Map.Line) elem );
                                }
                                else if ( elem instanceof Map.Sector ) {
                                        _map.flipSectorLines( (Map.Sector) elem );
                                }
                                else if ( elem instanceof Map.Vertex ) {
                                        _map.flipVertexLines( (Map.Vertex) elem );
                                }
                        }
                };
        
        private Map.ElementAction _splitLinesAction = new Map.ElementAction() {
                        public void doAction( Map.Element elem ) {
                                if ( elem instanceof Map.Line ) {
                                        _map.splitLine( (Map.Line) elem );
                                }
                        }
                };

        private static final String SELECT_CMD       = "Select";
        private static final String NEW_SECTOR_CMD   = "NewSect";
        private static final String NEW_LIGHT_CMD    = "NewLight";
        private static final String SNAP_CMD         = "Snap";
        private static final String CLEAR_CMD        = "Clear";
        private static final String GRID_SMALLER_CMD = "GridSmaller";
        private static final String GRID_LARGER_CMD  = "GridLarger";
        private static final String COPY_CMD         = "Copy";
        private static final String CUT_CMD          = "Cut";
        private static final String PASTE_CMD        = "Paste";
        private static final String MERGEVERTS_CMD   = "MergeVerts";
        private static final String FLIPLINE_CMD     = "FlipLine";
        private static final String SPLITLINE_CMD    = "SplitLine";
        private static final String UNSELECT_CMD     = "Unselect";
        private static final String STAIRS_CMD       = "MakeStairs";
        private static final String SET_CLIP_CMD     = "SetClip";
        private static final String EXPAND_CLIP_CMD  = "ExpandClip";
        private static final String SHRINK_CLIP_CMD  = "ShrinkClip";

        private Cursor _viewCursor = null;
        
        VCPairAction _cursorAction = new VCPairAction() {
                        public void doAction( VCPair vcp ) {
                                vcp._view.setCursor( _viewCursor );
                        }
                };
        
        public void setViewCursors( Cursor cursor ) {
                _viewCursor = cursor;
                performVCPairActions( _cursorAction );
        }
                                                        
        ActionListener menuListener = new ActionListener() {
                        public void actionPerformed( ActionEvent evt ) {
                                Object source = evt.getSource();
                                String command = evt.getActionCommand();
                                
                                if ( source == _snapItem ) {
                                        _infoBar._snappedField.setEnabled( _snapItem.isSelected() );
                                        _infoBar._gridSizeField.setText( "" + _gridSize + 
                                                                         ( _snapItem.isSelected() ? " (S)" : "" ) );
                                }
                                else if ( source == _clearItem || command.equals( CLEAR_CMD ) ) {
                                        cancelNewSector();
                                        _map.clear();
                                }
                                else if ( source == _gridSmallerItem || command.equals( GRID_SMALLER_CMD ) ) {
                                        setGridSize( _gridSize / 2 );
                                        // TODO: update some info bar/panel
                                }
                                else if ( source == _gridLargerItem || command.equals( GRID_LARGER_CMD ) ) {
                                        setGridSize( _gridSize * 2 );
                                        // TODO: update some info bar/panel
                                }
                                else if ( source == _selectItem ) {
                                        cancelNewSector();
                                        setViewCursors( null );
                                        _infoBar._modeField.setText( "Select" );
                                }
                                else if ( source == _clipItem ) {
                                        setViewCursors( Cursor.getPredefinedCursor( Cursor.HAND_CURSOR )  );
                                        _infoBar._modeField.setText( "Set Clip" );
                                }
                                else if ( source == _newSectorItem ) {
                                        setViewCursors( Cursor.getPredefinedCursor( Cursor.CROSSHAIR_CURSOR ) );
                                        _infoBar._modeField.setText( "New Sect" );
                                }
                                else if ( source == _newLightItem ) {
                                        setViewCursors( Cursor.getPredefinedCursor( Cursor.CROSSHAIR_CURSOR ) );
                                        _infoBar._modeField.setText( "New Light" );
                                }
                                else if ( source == _mergeVertsItem || command.equals( MERGEVERTS_CMD ) ) {
                                        Map.Vertex v = _map.getLastSelectedVertex();
                                        if ( v != null )
                                                _map.mergeVertexes( v );
                                }
                                else if ( source == _flipLineItem || command.equals( FLIPLINE_CMD ) ) {
                                        _map.doSelectionActions( _flipAction );
                                }
                                else if ( source == _splitLineItem || command.equals( SPLITLINE_CMD ) ) {
                                        _map.doSelectionActions( _splitLinesAction );
                                }
                                else if ( source == _stairsItem || command.equals( STAIRS_CMD ) ) {

                                        if ( ! _stairsItem.isEnabled() ) {
                                                Toolkit.getDefaultToolkit().beep();
                                                return;
                                        }

                                        Component comp = _menu.getComponent();
                                        Container cont = comp.getParent();

                                        while ( cont.getParent() != null ) {
                                                cont = cont.getParent();
                                                if ( cont instanceof Frame )
                                                        break;
                                        }
                                        
                                        MapUtil.showStairsDialog( ( cont instanceof Frame ) ?
                                                                  ( (Frame) cont ) : null , 
                                                                  _map );
                                }
                                else if ( source == _shrinkClipItem || command.equals( SHRINK_CLIP_CMD ) ) {
                                        Rectangle rect = _map.findSelectionsBounding();
                                        if ( rect == null ) {
                                                Toolkit.getDefaultToolkit().beep();
                                        }
                                        else {
                                                rect.grow( 16, 16 );
                                                _map.setUserClip( rect );
                                        }
                                }
                                else if ( source == _expandClipItem || command.equals( EXPAND_CLIP_CMD ) ) {
                                        _map.setUserClip( _map._mapRect );
                                }

                                else if ( source == _selectAllItem ) {
                                        // TODO: select all
                                }
                                else if ( source == _unselectItem || command.equals( UNSELECT_CMD ) ) {
                                        _map.clearSelections( false );

                                        if ( _newSectorItem.isSelected() )
                                                cancelNewSector();
                                }
                                else if ( command.equals( SELECT_CMD ) ) {
                                        _selectItem.doClick();
                                }
                                else if ( command.equals( NEW_LIGHT_CMD ) ) {
                                        _newLightItem.doClick();
                                }
                                else if ( command.equals( NEW_SECTOR_CMD ) ) {
                                        _newSectorItem.doClick();
                                }
                                else if ( command.equals( SET_CLIP_CMD ) ) {
                                        _clipItem.doClick();
                                }
                                else if ( command.equals( SNAP_CMD ) ) {
                                        _snapItem.doClick();
                                }
                        }     
                };

        private Map.Element clipboardElement = null;
        
        private ActionListener clipboardListener = new ActionListener() {
                        public void actionPerformed( ActionEvent evt ) {
                                Object source = evt.getSource();
                                String command = evt.getActionCommand();

                                if ( ( source == _copy || command.equals( COPY_CMD ) ) &&
                                     _map.hasSelections() ) {
                                        clipboardElement = _map.getLastSelection();
                                        _paste.setEnabled( true );
                                }

                                if ( ( source == _cut || command.equals( CUT_CMD ) ) && 
                                     _map.hasSelections() ) {
                                        clipboardElement =_map.getLastSelection();
                                        _map.removeElement( clipboardElement );
                                        _paste.setEnabled( true );
                                }
                                
                                if ( ( source == _paste || command.equals( PASTE_CMD ) ) && 
                                     clipboardElement != null ) {
                                        _paste.setEnabled( false );
                                }
                        }
                };

 
        public void registerViewController( View2d view, Controller2d controller ) {
                
                _vcPairs.addElement( new VCPair( view, controller ) );
                
                controller.setView( this, view );

                view.setController( this, controller );
                view.addMouseListener( controller );
                view.addMouseMotionListener( controller );

                view.setCursor( Cursor.getPredefinedCursor( Cursor.CROSSHAIR_CURSOR ) );
                
                view.registerKeyboardAction( menuListener,
                                             SELECT_CMD,
                                             KeyStroke.getKeyStroke( KeyEvent.VK_S, 0 ),
                                             JComponent.WHEN_FOCUSED );
                view.registerKeyboardAction( menuListener,
                                             NEW_SECTOR_CMD,
                                             KeyStroke.getKeyStroke( KeyEvent.VK_A, 0 ),
                                             JComponent.WHEN_FOCUSED );
                view.registerKeyboardAction( menuListener,
                                             NEW_LIGHT_CMD,
                                             KeyStroke.getKeyStroke( KeyEvent.VK_F, 0 ),
                                             JComponent.WHEN_FOCUSED );
                view.registerKeyboardAction( menuListener,
                                             SET_CLIP_CMD,
                                             KeyStroke.getKeyStroke( KeyEvent.VK_C, 0 ),
                                             JComponent.WHEN_FOCUSED );
                view.registerKeyboardAction( menuListener,
                                             SNAP_CMD,
                                             KeyStroke.getKeyStroke( KeyEvent.VK_G, Event.CTRL_MASK ),
                                             JComponent.WHEN_FOCUSED );
                view.registerKeyboardAction( menuListener,
                                             CLEAR_CMD,
                                             KeyStroke.getKeyStroke( KeyEvent.VK_K, Event.CTRL_MASK ),
                                             JComponent.WHEN_FOCUSED );
                view.registerKeyboardAction( menuListener,
                                             GRID_SMALLER_CMD,
                                             KeyStroke.getKeyStroke( KeyEvent.VK_OPEN_BRACKET, 0 ),
                                             JComponent.WHEN_FOCUSED );
                view.registerKeyboardAction( menuListener,
                                             GRID_LARGER_CMD,
                                             KeyStroke.getKeyStroke( KeyEvent.VK_CLOSE_BRACKET, 0 ),
                                             JComponent.WHEN_FOCUSED );
                view.registerKeyboardAction( clipboardListener,
                                             COPY_CMD,
                                             KeyStroke.getKeyStroke( KeyEvent.VK_C, Event.CTRL_MASK ),
                                             JComponent.WHEN_FOCUSED );
                view.registerKeyboardAction( clipboardListener,
                                             CUT_CMD,
                                             KeyStroke.getKeyStroke( KeyEvent.VK_X, Event.CTRL_MASK ),
                                             JComponent.WHEN_FOCUSED );
                view.registerKeyboardAction( clipboardListener,
                                             PASTE_CMD,
                                             KeyStroke.getKeyStroke( KeyEvent.VK_V, Event.CTRL_MASK ),
                                             JComponent.WHEN_FOCUSED );
                view.registerKeyboardAction( menuListener,
                                             MERGEVERTS_CMD,
                                             KeyStroke.getKeyStroke( KeyEvent.VK_M, Event.CTRL_MASK ),
                                             JComponent.WHEN_FOCUSED );
                view.registerKeyboardAction( menuListener,
                                             FLIPLINE_CMD,
                                             KeyStroke.getKeyStroke( KeyEvent.VK_F, Event.CTRL_MASK ),
                                             JComponent.WHEN_FOCUSED );
                view.registerKeyboardAction( menuListener,
                                             SPLITLINE_CMD,
                                             KeyStroke.getKeyStroke( KeyEvent.VK_I, Event.CTRL_MASK ),
                                             JComponent.WHEN_FOCUSED );
                view.registerKeyboardAction( menuListener,
                                             STAIRS_CMD,
                                             KeyStroke.getKeyStroke( KeyEvent.VK_T, Event.CTRL_MASK ),
                                             JComponent.WHEN_FOCUSED );

                view.registerKeyboardAction( menuListener,
                                             SHRINK_CLIP_CMD,
                                             KeyStroke.getKeyStroke( KeyEvent.VK_W, Event.CTRL_MASK ),
                                             JComponent.WHEN_FOCUSED );
                view.registerKeyboardAction( menuListener,
                                             EXPAND_CLIP_CMD,
                                             KeyStroke.getKeyStroke( KeyEvent.VK_A, Event.CTRL_MASK ),
                                             JComponent.WHEN_FOCUSED );

                view.registerKeyboardAction( menuListener,
                                             UNSELECT_CMD,
                                             KeyStroke.getKeyStroke( KeyEvent.VK_ESCAPE, 0 ),
                                             JComponent.WHEN_FOCUSED );
                
                
        }
        
        private void performVCPairActions( VCPairAction action ) {
                for ( Iterator iter = _vcPairs.iterator(); iter.hasNext(); ) {
                        action.doAction( (VCPair) iter.next() );
                }
        }
        
        private interface VCPairAction {
                public void doAction( VCPair vcp );
        }

        private class VCPair {
                View2d       _view       = null;
                Controller2d _controller = null;
                
                VCPair( View2d view, Controller2d controller ) {
                        _view = view;
                        _controller = controller;
                }
        }

        void cancelNewSector() {
                if ( _makingNewSector ) {
                        _map.numPtsInProg = 0;
                        _makingNewSector = false;
                        _map.setTransient( false, null, null );
                        _map.notifyListeners( null );
                }
        }

        void startNewSector( Point p ) {
                _makingNewSector = true;
                _map.addInProgress( p );
                _map.setTransient( true, p, p );
                _map.notifyListeners( null );
        }

        boolean isMakingNewSector() {
                return _makingNewSector;
        }

        /**
         * infobar shows:
         * - mode
         * - grid size (snap)
         * - zoom factor
         * - mouse coord
         */
        private class InfoBar extends JPanel {
                
                JTextField _modeField     = new JTextField( 9 );
                JTextField _gridSizeField = new JTextField( 6 );
                JTextField _zoomField     = new JTextField( 4 );
                JTextField _mouseField    = new JTextField( 12 );
                JTextField _snappedField  = new JTextField( 12 );
                
                private InfoBar() {
                        
                        _modeField.setEditable(     false );
                        _gridSizeField.setEditable( false );
                        _zoomField.setEditable(     false );
                        _mouseField.setEditable(    false );
                        _snappedField.setEditable(  false );
                        
                        _modeField.setBorder(     BorderFactory.createLoweredBevelBorder() );
                        _gridSizeField.setBorder( BorderFactory.createLoweredBevelBorder() );
                        _zoomField.setBorder(     BorderFactory.createLoweredBevelBorder() );
                        _mouseField.setBorder(    BorderFactory.createLoweredBevelBorder() );
                        _snappedField.setBorder(  BorderFactory.createLoweredBevelBorder() );
                        
                        _snappedField.setEnabled( _snapItem.isSelected() );

                        setLayout( new GridBagLayout() );
                        GridBagConstraints gbc = new GridBagConstraints();
                        
                        gbc.gridx = gbc.gridy = 0;
                        gbc.weighty = 0;
                        
                        gbc.fill = gbc.HORIZONTAL;
                        gbc.weightx = 0;
                        add( new JLabel( "Mode: " ), gbc );

                        ++gbc.gridx;
                        gbc.weightx = 1;
                        add( _modeField, gbc );

                        ++gbc.gridx;
                        gbc.weightx = 0;
                        add( new JLabel( "Grid: " ), gbc );

                        ++gbc.gridx;
                        gbc.weightx = 1;
                        add( _gridSizeField, gbc );

                        ++gbc.gridx;
                        gbc.weightx = 0;
                        add( new JLabel( "Zoom: " ), gbc );

                        ++gbc.gridx;
                        gbc.weightx = 1;
                        add( _zoomField, gbc );

                        ++gbc.gridx;
                        gbc.weightx = 0;
                        add( new JLabel( "Cursor: " ), gbc );

                        ++gbc.gridx;
                        gbc.weightx = 1;
                        add( _mouseField, gbc );

                        ++gbc.gridx;
                        gbc.weightx = 0;
                        add( new JLabel( "Snapped: " ), gbc );

                        ++gbc.gridx;
                        gbc.weightx = 1;
                        add( _snappedField, gbc );
                        
                        // add filler

                        ++gbc.gridx;
                        add( new JPanel(), gbc );

                }

        }
        
        private InfoBar _infoBar = new InfoBar();
        
        public JPanel getInfoBar() {
                return _infoBar;
        }

        void setInfoPointer( Point3D p ) {

                _infoBar._mouseField.setText( "" + p.x + ", " + p.y + ", " + p.z );
                Point3D sp = ( _snapItem.isSelected() ? getSnapped( p ) : p );
                _infoBar._snappedField.setText( "" + sp.x + ", " + sp.y + ", " + sp.z );
                boolean contained = _map.mapContains( p );
                _infoBar._mouseField.setEnabled( contained );                
                _infoBar._snappedField.setEnabled( contained && _snapItem.isSelected() );
        }
        
        void setInfoCurrent( View2d v, Controller2d c ) {
                if ( v != null )
                        setInfoZoom( v.getZoom() );
                else {
                        _infoBar._zoomField.setText(     "" );
                        _infoBar._mouseField.setText(    "" );
                        _infoBar._snappedField.setText(  "" );

                        _infoBar._mouseField.setEnabled(    false );
                        _infoBar._snappedField.setEnabled(  false );

                }
        }

        void setInfoZoom( float zoom ) {
                _infoBar._zoomField.setText( "" + (int) ( zoom * 100 ) + " %"  );
        }

        public void mapChanged( Rectangle clip ) {
                _copy.setEnabled(           _map.hasSelections() );
                _cut.setEnabled(            _map.hasSelections() );
                _mergeVertsItem.setEnabled( _map.hasTypeSelected( Map.Vertex.class ) );
                _flipLineItem.setEnabled(   _map.hasSelections() );
                _splitLineItem.setEnabled(  _map.hasTypeSelected( Map.Line.class ) );
                _stairsItem.setEnabled(     _map.getNumSelections() > 1 &&
                                            _map.hasTypeSelected( Map.Sector.class ) );
                _unselectItem.setEnabled(   _map.hasSelections() );

        }
        public void mapTransientChanged() {
                
        }

        public void mapSizeChanged() {
        }
}


